import textwrap
import uuid

import charmhelpers.contrib.network.ovs.ovn as ovn

import tests.utils as test_utils


CLUSTER_STATUS = textwrap.dedent("""
    0ea6
    Name: OVN_Northbound
    Cluster ID: f6a3 (f6a36e77-97bf-4740-b46a-705cbe4fef45)
    Server ID: 0ea6 (0ea6e785-c2bb-4640-b7a2-85104c11a2c1)
    Address: ssl:10.219.3.174:6643
    Status: cluster member
    Role: follower
    Term: 3
    Leader: 22dd
    Vote: unknown

    Election timer: 1000
    Log: [2, 10]
    Entries not yet committed: 0
    Entries not yet applied: 0
    Connections: ->f6cf ->22dd <-22dd <-f6cf
    Servers:
        0ea6 (0ea6 at ssl:10.219.3.174:6643) (self)
        f6cf (f6cf at ssl:10.219.3.64:6643)
        22dd (22dd at ssl:10.219.3.137:6643)
    """)

NORTHD_STATUS_ACTIVE = textwrap.dedent("""
    Status: active
    """)

NORTHD_STATUS_STANDBY = textwrap.dedent("""
    Status: standby
    """)


class TestOVN(test_utils.BaseTestCase):

    def test_ovn_appctl(self):
        self.patch_object(ovn.utils, '_run')
        ovn.ovn_appctl('ovn-northd', ('is-paused',))
        self._run.assert_called_once_with('ovn-appctl', '-t', 'ovn-northd',
                                          'is-paused')
        self._run.reset_mock()
        ovn.ovn_appctl('ovnnb_db', ('cluster/status',))
        self._run.assert_called_once_with('ovn-appctl', '-t',
                                          '/var/run/ovn/ovnnb_db.ctl',
                                          'cluster/status')
        self._run.reset_mock()
        ovn.ovn_appctl('ovnnb_db', ('cluster/status',), use_ovs_appctl=True)
        self._run.assert_called_once_with('ovs-appctl', '-t',
                                          '/var/run/ovn/ovnnb_db.ctl',
                                          'cluster/status')
        self._run.reset_mock()
        ovn.ovn_appctl('ovnsb_db', ('cluster/status',),
                       rundir='/var/run/openvswitch')
        self._run.assert_called_once_with('ovn-appctl', '-t',
                                          '/var/run/openvswitch/ovnsb_db.ctl',
                                          'cluster/status')

    def test_cluster_status(self):
        self.patch_object(ovn, 'ovn_appctl')
        self.ovn_appctl.return_value = CLUSTER_STATUS
        expect = ovn.OVNClusterStatus(
            'OVN_Northbound',
            uuid.UUID('f6a36e77-97bf-4740-b46a-705cbe4fef45'),
            uuid.UUID('0ea6e785-c2bb-4640-b7a2-85104c11a2c1'),
            'ssl:10.219.3.174:6643',
            'cluster member',
            'follower',
            3,
            '22dd',
            'unknown',
            1000,
            '[2, 10]',
            0,
            0,
            '->f6cf ->22dd <-22dd <-f6cf',
            [
                ('0ea6', 'ssl:10.219.3.174:6643'),
                ('f6cf', 'ssl:10.219.3.64:6643'),
                ('22dd', 'ssl:10.219.3.137:6643'),
            ])
        self.assertEquals(ovn.cluster_status('ovnnb_db'), expect)
        self.ovn_appctl.assert_called_once_with('ovnnb_db', ('cluster/status',
                                                'OVN_Northbound'),
                                                rundir=None,
                                                use_ovs_appctl=False)
        self.assertFalse(expect.is_cluster_leader)
        expect = ovn.OVNClusterStatus(
            'OVN_Northbound',
            uuid.UUID('f6a36e77-97bf-4740-b46a-705cbe4fef45'),
            uuid.UUID('0ea6e785-c2bb-4640-b7a2-85104c11a2c1'),
            'ssl:10.219.3.174:6643',
            'cluster member',
            'leader',
            3,
            'self',
            'unknown',
            1000,
            '[2, 10]',
            0,
            0,
            '->f6cf ->22dd <-22dd <-f6cf',
            [
                ('0ea6', 'ssl:10.219.3.174:6643'),
                ('f6cf', 'ssl:10.219.3.64:6643'),
                ('22dd', 'ssl:10.219.3.137:6643'),
            ])
        self.assertTrue(expect.is_cluster_leader)

    def test_is_northd_active(self):
        self.patch_object(ovn, 'ovn_appctl')
        self.ovn_appctl.return_value = NORTHD_STATUS_ACTIVE
        self.assertTrue(ovn.is_northd_active())
        self.ovn_appctl.assert_called_once_with('ovn-northd', ('status',))
        self.ovn_appctl.return_value = NORTHD_STATUS_STANDBY
        self.assertFalse(ovn.is_northd_active())



# Copyright 2014-2015 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest
from mock import patch, sentinel

from charmhelpers import context
from charmhelpers.core import hookenv


class TestRelations(unittest.TestCase):
    def setUp(self):
        def install(*args, **kw):
            p = patch.object(*args, **kw)
            p.start()
            self.addCleanup(p.stop)

        install(hookenv, 'relation_types', return_value=['rel', 'pear'])
        install(hookenv, 'peer_relation_id', return_value='pear:9')
        install(hookenv, 'relation_ids',
                side_effect=lambda x: ['{}:{}'.format(x, i)
                                       for i in range(9, 11)])
        install(hookenv, 'related_units',
                side_effect=lambda x: ['svc_' + x.replace(':', '/')])
        install(hookenv, 'local_unit', return_value='foo/1')
        install(hookenv, 'relation_get')
        install(hookenv, 'relation_set')
        # install(hookenv, 'is_leader', return_value=False)

    def test_relations(self):
        rels = context.Relations()
        self.assertListEqual(list(rels.keys()),
                             ['pear', 'rel'])  # Ordered alphabetically
        self.assertListEqual(list(rels['rel'].keys()),
                             ['rel:9', 'rel:10'])  # Ordered numerically

        # Relation data is loaded on demand, not on instantiation.
        self.assertFalse(hookenv.relation_get.called)

        # But we did have to retrieve some lists of units etc.
        self.assertGreaterEqual(hookenv.relation_ids.call_count, 2)
        self.assertGreaterEqual(hookenv.related_units.call_count, 2)

    def test_relations_peer(self):
        # The Relations instance has a short cut to the peer relation.
        # If the charm has managed to get multiple peer relations,
        # it returns the 'primary' one used here and returned by
        # hookenv.peer_relation_id()
        rels = context.Relations()
        self.assertIs(rels.peer, rels['pear']['pear:9'])

    def test_relation(self):
        rel = context.Relations()['rel']['rel:9']
        self.assertEqual(rel.relid, 'rel:9')
        self.assertEqual(rel.relname, 'rel')
        self.assertEqual(rel.service, 'svc_rel')
        self.assertTrue(isinstance(rel.local, context.RelationInfo))
        self.assertEqual(rel.local.unit, hookenv.local_unit())
        self.assertTrue(isinstance(rel.peers, context.OrderedDict))
        self.assertTrue(len(rel.peers), 2)
        self.assertTrue(isinstance(rel.peers['svc_pear/9'],
                                   context.RelationInfo))

        # I use this in my log messages. Relation id for identity
        # plus service name for ease of reference.
        self.assertEqual(str(rel), 'rel:9 (svc_rel)')

    def test_relation_no_peer_relation(self):
        hookenv.peer_relation_id.return_value = None
        rel = context.Relation('rel:10')
        self.assertTrue(rel.peers is None)

    def test_relation_no_peers(self):
        hookenv.related_units.side_effect = None
        hookenv.related_units.return_value = []
        rel = context.Relation('rel:10')
        self.assertDictEqual(rel.peers, {})

    def test_peer_relation(self):
        peer_rel = context.Relations().peer
        # The peer relation does not have a 'peers' properly. We
        # could give it one for symmetry, but it seems somewhat silly.
        self.assertTrue(peer_rel.peers is None)

    def test_relationinfo(self):
        hookenv.relation_get.return_value = {sentinel.key: 'value'}
        r = context.RelationInfo('rel:10', 'svc_rel/9')

        self.assertEqual(r.relname, 'rel')
        self.assertEqual(r.relid, 'rel:10')
        self.assertEqual(r.unit, 'svc_rel/9')
        self.assertEqual(r.service, 'svc_rel')
        self.assertEqual(r.number, 9)

        self.assertFalse(hookenv.relation_get.called)
        self.assertEqual(r[sentinel.key], 'value')
        hookenv.relation_get.assert_called_with(unit='svc_rel/9', rid='rel:10')

        # Updates fail
        with self.assertRaises(TypeError):
            r['newkey'] = 'foo'

        # Deletes fail
        with self.assertRaises(TypeError):
            del r[sentinel.key]

        # I use this for logging.
        self.assertEqual(str(r), 'rel:10 (svc_rel/9)')

    def test_relationinfo_local(self):
        r = context.RelationInfo('rel:10', hookenv.local_unit())

        # Updates work, with standard strings.
        r[sentinel.key] = 'value'
        hookenv.relation_set.assert_called_once_with(
            'rel:10', {sentinel.key: 'value'})

        # Python 2 unicode strings work too.
        hookenv.relation_set.reset_mock()
        r[sentinel.key] = 'value'
        hookenv.relation_set.assert_called_once_with(
            'rel:10', {sentinel.key: 'value'})

        # Byte strings fail under Python 3.
        with self.assertRaises(ValueError):
            r[sentinel.key] = b'value'

        # Deletes work
        del r[sentinel.key]
        hookenv.relation_set.assert_called_with('rel:10', {sentinel.key: None})

        # Attempting to write a non-string fails
        with self.assertRaises(ValueError):
            r[sentinel.key] = 42


class TestLeader(unittest.TestCase):
    @patch.object(hookenv, 'leader_get')
    def test_get(self, leader_get):
        leader_get.return_value = {'a_key': 'a_value'}

        leader = context.Leader()
        self.assertEqual(leader['a_key'], 'a_value')
        leader_get.assert_called_with()

        with self.assertRaises(KeyError):
            leader['missing']

    @patch.object(hookenv, 'leader_set')
    @patch.object(hookenv, 'leader_get')
    @patch.object(hookenv, 'is_leader')
    def test_set(self, is_leader, leader_get, leader_set):
        is_leader.return_value = True
        leader = context.Leader()

        # Updates work
        leader[sentinel.key] = 'foo'
        leader_set.assert_called_with({sentinel.key: 'foo'})
        del leader[sentinel.key]
        leader_set.assert_called_with({sentinel.key: None})

        # Python 2 unicode string values work too
        leader[sentinel.key] = 'bar'
        leader_set.assert_called_with({sentinel.key: 'bar'})

        # Byte strings fail under Python 3
        with self.assertRaises(ValueError):
            leader[sentinel.key] = b'baz'

        # Non strings fail, as implicit casting causes more trouble
        # than it solves. Simple types like integers would round trip
        # back as strings.
        with self.assertRaises(ValueError):
            leader[sentinel.key] = 42

    @patch.object(hookenv, 'leader_set')
    @patch.object(hookenv, 'leader_get')
    @patch.object(hookenv, 'is_leader')
    def test_set_not_leader(self, is_leader, leader_get, leader_set):
        is_leader.return_value = False
        leader_get.return_value = {'a_key': 'a_value'}
        leader = context.Leader()
        with self.assertRaises(TypeError):
            leader['a_key'] = 'foo'
        with self.assertRaises(TypeError):
            del leader['a_key']



import copy
import json

from tests.helpers import FakeRelation
from testtools import TestCase
from mock import patch, call
from charmhelpers.contrib import peerstorage


TO_PATCH = [
    'current_relation_id',
    'is_relation_made',
    'local_unit',
    'relation_get',
    '_relation_get',
    'relation_ids',
    'relation_set',
    '_relation_set',
    '_leader_get',
    'leader_set',
    'is_leader',
]
FAKE_RELATION_NAME = 'cluster'
FAKE_RELATION = {
    'cluster:0': {
        'cluster/0': {
        },
        'cluster/1': {
        },
        'cluster/2': {
        },
    },

}
FAKE_RELATION_IDS = ['cluster:0']
FAKE_LOCAL_UNIT = 'test_host'


class TestPeerStorage(TestCase):
    def setUp(self):
        super(TestPeerStorage, self).setUp()
        for m in TO_PATCH:
            setattr(self, m, self._patch(m))
        self.fake_relation_name = FAKE_RELATION_NAME
        self.fake_relation = FakeRelation(FAKE_RELATION)
        self.local_unit.return_value = FAKE_LOCAL_UNIT
        self.relation_get.return_value = {'key1': 'value1',
                                          'key2': 'value2',
                                          'private-address': '127.0.0.1',
                                          'public-address': '91.189.90.159'}

    def _patch(self, method):
        _m = patch('charmhelpers.contrib.peerstorage.' + method)
        mock = _m.start()
        self.addCleanup(_m.stop)
        return mock

    def test_peer_retrieve_no_relation(self):
        self.relation_ids.return_value = []
        self.assertRaises(ValueError, peerstorage.peer_retrieve, 'key', relation_name=self.fake_relation_name)

    def test_peer_retrieve_with_relation(self):
        self.relation_ids.return_value = FAKE_RELATION_IDS
        peerstorage.peer_retrieve('key', self.fake_relation_name)
        self.relation_get.assert_called_with(attribute='key', rid=FAKE_RELATION_IDS[0], unit=FAKE_LOCAL_UNIT)

    def test_peer_store_no_relation(self):
        self.relation_ids.return_value = []
        self.assertRaises(ValueError, peerstorage.peer_store, 'key', 'value', relation_name=self.fake_relation_name)

    def test_peer_store_with_relation(self):
        self.relation_ids.return_value = FAKE_RELATION_IDS
        peerstorage.peer_store('key', 'value', self.fake_relation_name)
        self.relation_set.assert_called_with(relation_id=FAKE_RELATION_IDS[0],
                                             relation_settings={'key': 'value'})

    def test_peer_echo_no_includes(self):
        peerstorage.is_leader.side_effect = NotImplementedError
        settings = {'key1': 'value1', 'key2': 'value2'}
        self._relation_get.copy.return_value = settings
        self._relation_get.return_value = settings
        peerstorage.peer_echo()
        self._relation_set.assert_called_with(relation_settings=settings)

    def test_peer_echo_includes(self):
        peerstorage.is_leader.side_effect = NotImplementedError
        settings = {'key1': 'value1'}
        self._relation_get.copy.return_value = settings
        self._relation_get.return_value = settings
        peerstorage.peer_echo(['key1'])
        self._relation_set.assert_called_with(relation_settings=settings)

    @patch.object(peerstorage, 'peer_store')
    def test_peer_store_and_set_no_relation(self, peer_store):
        self.is_relation_made.return_value = False
        peerstorage.peer_store_and_set(relation_id='db', kwarg1='kwarg1_v')
        self.relation_set.assert_called_with(relation_id='db',
                                             relation_settings={},
                                             kwarg1='kwarg1_v')
        peer_store.assert_not_called()

    @patch.object(peerstorage, 'peer_store')
    def test_peer_store_and_set_no_relation_fatal(self, peer_store):
        self.is_relation_made.return_value = False
        self.assertRaises(ValueError,
                          peerstorage.peer_store_and_set,
                          relation_id='db',
                          kwarg1='kwarg1_v',
                          peer_store_fatal=True)

    @patch.object(peerstorage, 'peer_store')
    def test_peer_store_and_set_kwargs(self, peer_store):
        self.is_relation_made.return_value = True
        peerstorage.peer_store_and_set(relation_id='db', kwarg1='kwarg1_v')
        self.relation_set.assert_called_with(relation_id='db',
                                             relation_settings={},
                                             kwarg1='kwarg1_v')
        calls = [call('db_kwarg1', 'kwarg1_v', relation_name='cluster')]
        peer_store.assert_has_calls(calls, any_order=True)

    @patch.object(peerstorage, 'peer_store')
    def test_peer_store_and_rel_settings(self, peer_store):
        self.is_relation_made.return_value = True
        rel_setting = {
            'rel_set1': 'relset1_v'
        }
        peerstorage.peer_store_and_set(relation_id='db',
                                       relation_settings=rel_setting)
        self.relation_set.assert_called_with(relation_id='db',
                                             relation_settings=rel_setting)
        calls = [call('db_rel_set1', 'relset1_v', relation_name='cluster')]
        peer_store.assert_has_calls(calls, any_order=True)

    @patch.object(peerstorage, 'peer_store')
    def test_peer_store_and_set(self, peer_store):
        self.is_relation_made.return_value = True
        rel_setting = {
            'rel_set1': 'relset1_v'
        }
        peerstorage.peer_store_and_set(relation_id='db',
                                       relation_settings=rel_setting,
                                       kwarg1='kwarg1_v',
                                       delimiter='+')
        self.relation_set.assert_called_with(relation_id='db',
                                             relation_settings=rel_setting,
                                             kwarg1='kwarg1_v')
        calls = [call('db+rel_set1', 'relset1_v', relation_name='cluster'),
                 call('db+kwarg1', 'kwarg1_v', relation_name='cluster')]
        peer_store.assert_has_calls(calls, any_order=True)

    @patch.object(peerstorage, 'peer_retrieve')
    def test_peer_retrieve_by_prefix(self, peer_retrieve):
        rel_id = 'db:2'
        settings = {
            'user': 'bob',
            'pass': 'reallyhardpassword',
            'host': 'myhost',
        }
        peer_settings = {rel_id + '_' + k: v for k, v in settings.items()}
        peer_retrieve.return_value = peer_settings
        self.assertEquals(peerstorage.peer_retrieve_by_prefix(rel_id), settings)

    @patch.object(peerstorage, 'peer_retrieve')
    def test_peer_retrieve_by_prefix_empty_relation(self, peer_retrieve):
        # If relation-get returns None, peer_retrieve_by_prefix returns
        # an empty dictionary.
        peer_retrieve.return_value = None
        rel_id = 'db:2'
        self.assertEquals(peerstorage.peer_retrieve_by_prefix(rel_id), {})

    @patch.object(peerstorage, 'peer_retrieve')
    def test_peer_retrieve_by_prefix_exc_list(self, peer_retrieve):
        rel_id = 'db:2'
        settings = {
            'user': 'bob',
            'pass': 'reallyhardpassword',
            'host': 'myhost',
        }
        peer_settings = {rel_id + '_' + k: v for k, v in settings.items()}
        del settings['host']
        peer_retrieve.return_value = peer_settings
        self.assertEquals(peerstorage.peer_retrieve_by_prefix(rel_id,
                                                              exc_list=['host']),
                          settings)

    @patch.object(peerstorage, 'peer_retrieve')
    def test_peer_retrieve_by_prefix_inc_list(self, peer_retrieve):
        rel_id = 'db:2'
        settings = {
            'user': 'bob',
            'pass': 'reallyhardpassword',
            'host': 'myhost',
        }
        peer_settings = {rel_id + '_' + k: v for k, v in settings.items()}
        peer_retrieve.return_value = peer_settings
        self.assertEquals(peerstorage.peer_retrieve_by_prefix(rel_id,
                                                              inc_list=['host']),
                          {'host': 'myhost'})

    def test_leader_get_migration_is_leader(self):
        self.is_leader.return_value = True
        l_settings = {'s3': 3}
        r_settings = {'s1': 1, 's2': 2}

        def mock_relation_get(attribute=None, unit=None, rid=None):
            if attribute:
                if attribute in r_settings:
                    return r_settings.get(attribute)
                else:
                    return None

            return copy.deepcopy(r_settings)

        def mock_leader_get(attribute=None):
            if attribute:
                if attribute in l_settings:
                    return l_settings.get(attribute)
                else:
                    return None

            return copy.deepcopy(l_settings)

        def mock_leader_set(settings=None, **kwargs):
            if settings:
                l_settings.update(settings)

            l_settings.update(kwargs)

        def check_leader_db(dicta, dictb):
            _dicta = copy.deepcopy(dicta)
            _dictb = copy.deepcopy(dictb)
            miga = json.loads(_dicta[migration_key]).sort()
            migb = json.loads(_dictb[migration_key]).sort()
            self.assertEqual(miga, migb)
            del _dicta[migration_key]
            del _dictb[migration_key]
            self.assertEqual(_dicta, _dictb)

        migration_key = '__leader_get_migrated_settings__'
        self._relation_get.side_effect = mock_relation_get
        self._leader_get.side_effect = mock_leader_get
        self.leader_set.side_effect = mock_leader_set

        self.assertEqual({'s1': 1, 's2': 2}, peerstorage._relation_get())
        self.assertEqual({'s3': 3}, peerstorage._leader_get())
        self.assertEqual({'s1': 1, 's2': 2, 's3': 3}, peerstorage.leader_get())
        check_leader_db({'s1': 1, 's2': 2, 's3': 3,
                         migration_key: '["s2", "s1"]'}, l_settings)
        self.assertTrue(peerstorage.leader_set.called)

        peerstorage.leader_set.reset_mock()
        self.assertEqual({'s1': 1, 's2': 2, 's3': 3}, peerstorage.leader_get())
        check_leader_db({'s1': 1, 's2': 2, 's3': 3,
                         migration_key: '["s2", "s1"]'}, l_settings)
        self.assertFalse(peerstorage.leader_set.called)

        l_settings = {'s3': 3}
        peerstorage.leader_set.reset_mock()
        self.assertEqual(1, peerstorage.leader_get('s1'))
        check_leader_db({'s1': 1, 's3': 3,
                         migration_key: '["s1"]'}, l_settings)
        self.assertTrue(peerstorage.leader_set.called)

        # Test that leader vals take precedence over non-leader vals
        r_settings['s3'] = 2
        r_settings['s4'] = 3
        l_settings['s4'] = 4

        peerstorage.leader_set.reset_mock()
        self.assertEqual(4, peerstorage.leader_get('s4'))
        check_leader_db({'s1': 1, 's3': 3, 's4': 4,
                         migration_key: '["s1", "s4"]'}, l_settings)
        self.assertTrue(peerstorage.leader_set.called)

        peerstorage.leader_set.reset_mock()
        self.assertEqual({'s1': 1, 's2': 2, 's3': 2, 's4': 3},
                         peerstorage._relation_get())
        check_leader_db({'s1': 1, 's3': 3, 's4': 4,
                         migration_key: '["s1", "s4"]'},
                        peerstorage._leader_get())
        self.assertEqual({'s1': 1, 's2': 2, 's3': 3, 's4': 4},
                         peerstorage.leader_get())
        check_leader_db({'s1': 1, 's2': 2, 's3': 3, 's4': 4,
                         migration_key: '["s3", "s2", "s1", "s4"]'},
                        l_settings)
        self.assertTrue(peerstorage.leader_set.called)

    def test_leader_get_migration_is_not_leader(self):
        self.is_leader.return_value = False
        l_settings = {'s3': 3}
        r_settings = {'s1': 1, 's2': 2}

        def mock_relation_get(attribute=None, unit=None, rid=None):
            if attribute:
                if attribute in r_settings:
                    return r_settings.get(attribute)
                else:
                    return None

            return copy.deepcopy(r_settings)

        def mock_leader_get(attribute=None):
            if attribute:
                if attribute in l_settings:
                    return l_settings.get(attribute)
                else:
                    return None

            return copy.deepcopy(l_settings)

        def mock_leader_set(settings=None, **kwargs):
            if settings:
                l_settings.update(settings)

            l_settings.update(kwargs)

        self._relation_get.side_effect = mock_relation_get
        self._leader_get.side_effect = mock_leader_get
        self.leader_set.side_effect = mock_leader_set
        self.assertEqual({'s1': 1, 's2': 2}, peerstorage._relation_get())
        self.assertEqual({'s3': 3}, peerstorage._leader_get())
        self.assertEqual({'s3': 3}, peerstorage.leader_get())
        self.assertEqual({'s3': 3}, l_settings)
        self.assertFalse(peerstorage.leader_set.called)

        self.assertEqual({'s3': 3}, peerstorage.leader_get())
        self.assertEqual({'s3': 3}, l_settings)
        self.assertFalse(peerstorage.leader_set.called)

        # Test that leader vals take precedence over non-leader vals
        r_settings['s3'] = 2
        r_settings['s4'] = 3
        l_settings['s4'] = 4

        self.assertEqual(4, peerstorage.leader_get('s4'))
        self.assertEqual({'s3': 3, 's4': 4}, l_settings)
        self.assertFalse(peerstorage.leader_set.called)



#!/usr/bin/env python
# coding: utf-8

from unittest import TestCase

from charmhelpers.fetch.python import debug
from charmhelpers.fetch.python import packages
from charmhelpers.fetch.python import rpdb
from charmhelpers.fetch.python import version
from charmhelpers.contrib.python import debug as contrib_debug
from charmhelpers.contrib.python import packages as contrib_packages
from charmhelpers.contrib.python import rpdb as contrib_rpdb
from charmhelpers.contrib.python import version as contrib_version


class ContribDebugTestCase(TestCase):
    def test_aliases(self):
        assert contrib_debug is debug
        assert contrib_packages is packages
        assert contrib_rpdb is rpdb
        assert contrib_version is version



import os
import yaml

from testtools import TestCase
from mock import (
    patch,
    MagicMock,
    call,
)

from charmhelpers import fetch

from urllib.parse import urlparse


FAKE_APT_CACHE = {
    # an installed package
    'vim': {
        'current_ver': '2:7.3.547-6ubuntu5'
    },
    # a uninstalled installation candidate
    'emacs': {
    }
}


def fake_apt_cache(in_memory=True, progress=None):
    def _get(package):
        pkg = MagicMock()
        if package not in FAKE_APT_CACHE:
            raise KeyError
        pkg.name = package
        if 'current_ver' in FAKE_APT_CACHE[package]:
            pkg.current_ver.ver_str = FAKE_APT_CACHE[package]['current_ver']
        else:
            pkg.current_ver = None
        return pkg
    cache = MagicMock()
    cache.__getitem__.side_effect = _get
    return cache


def getenv(update=None):
    # return a copy of os.environ with update applied.
    # this was necessary because some modules modify os.environment directly
    copy = os.environ.copy()
    if update is not None:
        copy.update(update)
    return copy


class FetchTest(TestCase):

    @patch('charmhelpers.fetch.log')
    @patch.object(fetch, 'config')
    @patch.object(fetch, 'add_source')
    def test_configure_sources_single_source(self, add_source, config, log):
        config.side_effect = ['source', 'key']
        fetch.configure_sources()
        add_source.assert_called_with('source', 'key')

    @patch.object(fetch, 'config')
    @patch.object(fetch, 'add_source')
    def test_configure_sources_null_source(self, add_source, config):
        config.side_effect = [None, None]
        fetch.configure_sources()
        self.assertEqual(add_source.call_count, 0)

    @patch.object(fetch, 'config')
    @patch.object(fetch, 'add_source')
    def test_configure_sources_empty_source(self, add_source, config):
        config.side_effect = ['', '']
        fetch.configure_sources()
        self.assertEqual(add_source.call_count, 0)

    @patch.object(fetch, 'config')
    @patch.object(fetch, 'add_source')
    def test_configure_sources_single_source_no_key(self, add_source, config):
        config.side_effect = ['source', None]
        fetch.configure_sources()
        add_source.assert_called_with('source', None)

    @patch.object(fetch, 'config')
    @patch.object(fetch, 'add_source')
    def test_configure_sources_multiple_sources(self, add_source, config):
        sources = ["sourcea", "sourceb"]
        keys = ["keya", None]
        config.side_effect = [
            yaml.dump(sources),
            yaml.dump(keys)
        ]
        fetch.configure_sources()
        add_source.assert_has_calls([
            call('sourcea', 'keya'),
            call('sourceb', None)
        ])

    @patch.object(fetch, 'config')
    @patch.object(fetch, 'add_source')
    def test_configure_sources_missing_keys(self, add_source, config):
        sources = ["sourcea", "sourceb"]
        keys = ["keya"]  # Second key is missing
        config.side_effect = [
            yaml.dump(sources),
            yaml.dump(keys)
        ]
        self.assertRaises(fetch.SourceConfigError, fetch.configure_sources)

    @patch.object(fetch, '_fetch_update')
    @patch.object(fetch, 'config')
    @patch.object(fetch, 'add_source')
    def test_configure_sources_update_called_ubuntu(self, add_source, config,
                                                    update):
        config.side_effect = ['source', 'key']
        fetch.configure_sources(update=True)
        add_source.assert_called_with('source', 'key')
        self.assertTrue(update.called)


class InstallTest(TestCase):

    def setUp(self):
        super(InstallTest, self).setUp()
        self.valid_urls = (
            "http://example.com/foo.tar.gz",
            "http://example.com/foo.tgz",
            "http://example.com/foo.tar.bz2",
            "http://example.com/foo.tbz2",
            "http://example.com/foo.zip",
            "http://example.com/foo.zip?bar=baz&x=y#whee",
            "ftp://example.com/foo.tar.gz",
            "https://example.com/foo.tgz",
            "file://example.com/foo.tar.bz2",
            "bzr+ssh://example.com/branch-name",
            "bzr+ssh://example.com/branch-name/",
            "lp:branch-name",
            "lp:example/branch-name",
        )
        self.invalid_urls = (
            "git://example.com/foo.tar.gz",
            "http://example.com/foo",
            "http://example.com/foobar=baz&x=y#tar.gz",
            "http://example.com/foobar?h=baz.zip",
            "abc:example",
            "file//example.com/foo.tar.bz2",
            "garbage",
        )

    @patch('charmhelpers.fetch.log')
    @patch('charmhelpers.fetch.plugins')
    def test_installs_remote(self, _plugins, _log):
        h1 = MagicMock(name="h1")
        h1.can_handle.return_value = "Nope"

        h2 = MagicMock(name="h2")
        h2.can_handle.return_value = True
        h2.install.side_effect = fetch.UnhandledSource()

        h3 = MagicMock(name="h3")
        h3.can_handle.return_value = True
        h3.install.return_value = "foo"

        _plugins.return_value = [h1, h2, h3]
        for url in self.valid_urls:
            result = fetch.install_remote(url)

            h1.can_handle.assert_called_with(url)
            h2.can_handle.assert_called_with(url)
            h3.can_handle.assert_called_with(url)

            h1.install.assert_not_called()
            h2.install.assert_called_with(url)
            h3.install.assert_called_with(url)

            self.assertEqual(result, "foo")

        fetch.install_remote('url', extra_arg=True)
        h2.install.assert_called_with('url', extra_arg=True)

    @patch('charmhelpers.fetch.install_remote')
    @patch('charmhelpers.fetch.config')
    def test_installs_from_config(self, _config, _instrem):
        for url in self.valid_urls:
            _config.return_value = {"foo": url}
            fetch.install_from_config("foo")
            _instrem.assert_called_with(url)


class PluginTest(TestCase):

    @patch('charmhelpers.fetch.importlib.import_module')
    def test_imports_plugins(self, import_):
        fetch_handlers = ['a.foo', 'b.foo', 'c.foo']
        module = MagicMock()
        import_.return_value = module
        plugins = fetch.plugins(fetch_handlers)

        self.assertEqual(len(fetch_handlers), len(plugins))
        module.foo.assert_has_calls(([call()] * len(fetch_handlers)))

    @patch('charmhelpers.fetch.importlib.import_module')
    def test_imports_plugins_default(self, import_):
        module = MagicMock()
        import_.return_value = module
        plugins = fetch.plugins()

        self.assertEqual(len(fetch.FETCH_HANDLERS), len(plugins))
        for handler in fetch.FETCH_HANDLERS:
            classname = handler.rsplit('.', 1)[-1]
            getattr(module, classname).assert_called_with()

    @patch('charmhelpers.fetch.log')
    @patch('charmhelpers.fetch.importlib.import_module')
    def test_skips_and_logs_missing_plugins(self, import_, log_):
        fetch_handlers = ['a.foo', 'b.foo', 'c.foo']
        import_.side_effect = (NotImplementedError, NotImplementedError,
                               MagicMock())
        plugins = fetch.plugins(fetch_handlers)

        self.assertEqual(1, len(plugins))
        self.assertEqual(2, log_.call_count)

    @patch('charmhelpers.fetch.log')
    @patch.object(fetch.importlib, 'import_module')
    def test_plugins_are_valid(self, import_module, log_):
        plugins = fetch.plugins()
        self.assertEqual(len(fetch.FETCH_HANDLERS), len(plugins))


class BaseFetchHandlerTest(TestCase):

    def setUp(self):
        super(BaseFetchHandlerTest, self).setUp()
        self.test_urls = (
            "http://example.com/foo?bar=baz&x=y#blarg",
            "https://example.com/foo",
            "ftp://example.com/foo",
            "file://example.com/foo",
            "git://github.com/foo/bar",
            "bzr+ssh://bazaar.launchpad.net/foo/bar",
            "bzr+http://bazaar.launchpad.net/foo/bar",
            "garbage",
        )
        self.fh = fetch.BaseFetchHandler()

    def test_handles_nothing(self):
        for url in self.test_urls:
            self.assertNotEqual(self.fh.can_handle(url), True)

    def test_install_throws_unhandled(self):
        for url in self.test_urls:
            self.assertRaises(fetch.UnhandledSource, self.fh.install, url)

    def test_parses_urls(self):
        sample_url = "http://example.com/foo?bar=baz&x=y#blarg"
        p = self.fh.parse_url(sample_url)
        self.assertEqual(p, urlparse(sample_url))

    def test_returns_baseurl(self):
        sample_url = "http://example.com/foo?bar=baz&x=y#blarg"
        expected_url = "http://example.com/foo"
        u = self.fh.base_url(sample_url)
        self.assertEqual(u, expected_url)



# Copyright 2014-2015 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import cmdline
from charmhelpers.contrib.benchmark import Benchmark


@cmdline.subcommand(command_name='benchmark-start')
def start():
    Benchmark.start()


@cmdline.subcommand(command_name='benchmark-finish')
def finish():
    Benchmark.finish()


@cmdline.subcommand_builder('benchmark-composite', description="Set the benchmark composite score")
def service(subparser):
    subparser.add_argument("value", help="The composite score.")
    subparser.add_argument("units", help="The units the composite score represents, i.e., 'reads/sec'.")
    subparser.add_argument("direction", help="'asc' if a lower score is better, 'desc' if a higher score is better.")
    return Benchmark.set_composite_score



# Copyright 2014-2015 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json
import inspect
from collections import Iterable, OrderedDict

from charmhelpers.core import host
from charmhelpers.core import hookenv


__all__ = ['ServiceManager', 'ManagerCallback',
           'PortManagerCallback', 'open_ports', 'close_ports', 'manage_ports',
           'service_restart', 'service_stop']


class ServiceManager(object):
    def __init__(self, services=None):
        """
        Register a list of services, given their definitions.

        Service definitions are dicts in the following formats (all keys except
        'service' are optional)::

            {
                "service": <service name>,
                "required_data": <list of required data contexts>,
                "provided_data": <list of provided data contexts>,
                "data_ready": <one or more callbacks>,
                "data_lost": <one or more callbacks>,
                "start": <one or more callbacks>,
                "stop": <one or more callbacks>,
                "ports": <list of ports to manage>,
            }

        The 'required_data' list should contain dicts of required data (or
        dependency managers that act like dicts and know how to collect the data).
        Only when all items in the 'required_data' list are populated are the list
        of 'data_ready' and 'start' callbacks executed.  See `is_ready()` for more
        information.

        The 'provided_data' list should contain relation data providers, most likely
        a subclass of :class:`charmhelpers.core.services.helpers.RelationContext`,
        that will indicate a set of data to set on a given relation.

        The 'data_ready' value should be either a single callback, or a list of
        callbacks, to be called when all items in 'required_data' pass `is_ready()`.
        Each callback will be called with the service name as the only parameter.
        After all of the 'data_ready' callbacks are called, the 'start' callbacks
        are fired.

        The 'data_lost' value should be either a single callback, or a list of
        callbacks, to be called when a 'required_data' item no longer passes
        `is_ready()`.  Each callback will be called with the service name as the
        only parameter.  After all of the 'data_lost' callbacks are called,
        the 'stop' callbacks are fired.

        The 'start' value should be either a single callback, or a list of
        callbacks, to be called when starting the service, after the 'data_ready'
        callbacks are complete.  Each callback will be called with the service
        name as the only parameter.  This defaults to
        `[host.service_start, services.open_ports]`.

        The 'stop' value should be either a single callback, or a list of
        callbacks, to be called when stopping the service.  If the service is
        being stopped because it no longer has all of its 'required_data', this
        will be called after all of the 'data_lost' callbacks are complete.
        Each callback will be called with the service name as the only parameter.
        This defaults to `[services.close_ports, host.service_stop]`.

        The 'ports' value should be a list of ports to manage.  The default
        'start' handler will open the ports after the service is started,
        and the default 'stop' handler will close the ports prior to stopping
        the service.


        Examples:

        The following registers an Upstart service called bingod that depends on
        a mongodb relation and which runs a custom `db_migrate` function prior to
        restarting the service, and a Runit service called spadesd::

            manager = services.ServiceManager([
                {
                    'service': 'bingod',
                    'ports': [80, 443],
                    'required_data': [MongoRelation(), config(), {'my': 'data'}],
                    'data_ready': [
                        services.template(source='bingod.conf'),
                        services.template(source='bingod.ini',
                                          target='/etc/bingod.ini',
                                          owner='bingo', perms=0400),
                    ],
                },
                {
                    'service': 'spadesd',
                    'data_ready': services.template(source='spadesd_run.j2',
                                                    target='/etc/sv/spadesd/run',
                                                    perms=0555),
                    'start': runit_start,
                    'stop': runit_stop,
                },
            ])
            manager.manage()
        """
        self._ready_file = os.path.join(hookenv.charm_dir(), 'READY-SERVICES.json')
        self._ready = None
        self.services = OrderedDict()
        for service in services or []:
            service_name = service['service']
            self.services[service_name] = service

    def manage(self):
        """
        Handle the current hook by doing The Right Thing with the registered services.
        """
        hookenv._run_atstart()
        try:
            hook_name = hookenv.hook_name()
            if hook_name == 'stop':
                self.stop_services()
            else:
                self.reconfigure_services()
                self.provide_data()
        except SystemExit as x:
            if x.code is None or x.code == 0:
                hookenv._run_atexit()
        hookenv._run_atexit()

    def provide_data(self):
        """
        Set the relation data for each provider in the ``provided_data`` list.

        A provider must have a `name` attribute, which indicates which relation
        to set data on, and a `provide_data()` method, which returns a dict of
        data to set.

        The `provide_data()` method can optionally accept two parameters:

          * ``remote_service`` The name of the remote service that the data will
            be provided to.  The `provide_data()` method will be called once
            for each connected service (not unit).  This allows the method to
            tailor its data to the given service.
          * ``service_ready`` Whether or not the service definition had all of
            its requirements met, and thus the ``data_ready`` callbacks run.

        Note that the ``provided_data`` methods are now called **after** the
        ``data_ready`` callbacks are run.  This gives the ``data_ready`` callbacks
        a chance to generate any data necessary for the providing to the remote
        services.
        """
        for service_name, service in self.services.items():
            service_ready = self.is_ready(service_name)
            for provider in service.get('provided_data', []):
                for relid in hookenv.relation_ids(provider.name):
                    units = hookenv.related_units(relid)
                    if not units:
                        continue
                    remote_service = units[0].split('/')[0]
                    argspec = inspect.getfullargspec(provider.provide_data)
                    if len(argspec.args) > 1:
                        data = provider.provide_data(remote_service, service_ready)
                    else:
                        data = provider.provide_data()
                    if data:
                        hookenv.relation_set(relid, data)

    def reconfigure_services(self, *service_names):
        """
        Update all files for one or more registered services, and,
        if ready, optionally restart them.

        If no service names are given, reconfigures all registered services.
        """
        for service_name in service_names or self.services.keys():
            if self.is_ready(service_name):
                self.fire_event('data_ready', service_name)
                self.fire_event('start', service_name, default=[
                    service_restart,
                    manage_ports])
                self.save_ready(service_name)
            else:
                if self.was_ready(service_name):
                    self.fire_event('data_lost', service_name)
                self.fire_event('stop', service_name, default=[
                    manage_ports,
                    service_stop])
                self.save_lost(service_name)

    def stop_services(self, *service_names):
        """
        Stop one or more registered services, by name.

        If no service names are given, stops all registered services.
        """
        for service_name in service_names or self.services.keys():
            self.fire_event('stop', service_name, default=[
                manage_ports,
                service_stop])

    def get_service(self, service_name):
        """
        Given the name of a registered service, return its service definition.
        """
        service = self.services.get(service_name)
        if not service:
            raise KeyError('Service not registered: %s' % service_name)
        return service

    def fire_event(self, event_name, service_name, default=None):
        """
        Fire a data_ready, data_lost, start, or stop event on a given service.
        """
        service = self.get_service(service_name)
        callbacks = service.get(event_name, default)
        if not callbacks:
            return
        if not isinstance(callbacks, Iterable):
            callbacks = [callbacks]
        for callback in callbacks:
            if isinstance(callback, ManagerCallback):
                callback(self, service_name, event_name)
            else:
                callback(service_name)

    def is_ready(self, service_name):
        """
        Determine if a registered service is ready, by checking its 'required_data'.

        A 'required_data' item can be any mapping type, and is considered ready
        if `bool(item)` evaluates as True.
        """
        service = self.get_service(service_name)
        reqs = service.get('required_data', [])
        return all(bool(req) for req in reqs)

    def _load_ready_file(self):
        if self._ready is not None:
            return
        if os.path.exists(self._ready_file):
            with open(self._ready_file) as fp:
                self._ready = set(json.load(fp))
        else:
            self._ready = set()

    def _save_ready_file(self):
        if self._ready is None:
            return
        with open(self._ready_file, 'w') as fp:
            json.dump(list(self._ready), fp)

    def save_ready(self, service_name):
        """
        Save an indicator that the given service is now data_ready.
        """
        self._load_ready_file()
        self._ready.add(service_name)
        self._save_ready_file()

    def save_lost(self, service_name):
        """
        Save an indicator that the given service is no longer data_ready.
        """
        self._load_ready_file()
        self._ready.discard(service_name)
        self._save_ready_file()

    def was_ready(self, service_name):
        """
        Determine if the given service was previously data_ready.
        """
        self._load_ready_file()
        return service_name in self._ready


class ManagerCallback(object):
    """
    Special case of a callback that takes the `ServiceManager` instance
    in addition to the service name.

    Subclasses should implement `__call__` which should accept three parameters:

        * `manager`       The `ServiceManager` instance
        * `service_name`  The name of the service it's being triggered for
        * `event_name`    The name of the event that this callback is handling
    """
    def __call__(self, manager, service_name, event_name):
        raise NotImplementedError()


class PortManagerCallback(ManagerCallback):
    """
    Callback class that will open or close ports, for use as either
    a start or stop action.
    """
    def __call__(self, manager, service_name, event_name):
        service = manager.get_service(service_name)
        # turn this generator into a list,
        # as we'll be going over it multiple times
        new_ports = list(service.get('ports', []))
        port_file = os.path.join(hookenv.charm_dir(), '.{}.ports'.format(service_name))
        if os.path.exists(port_file):
            with open(port_file) as fp:
                old_ports = fp.read().split(',')
            for old_port in old_ports:
                if bool(old_port) and not self.ports_contains(old_port, new_ports):
                    hookenv.close_port(old_port)
        with open(port_file, 'w') as fp:
            fp.write(','.join(str(port) for port in new_ports))
        for port in new_ports:
            # A port is either a number or 'ICMP'
            protocol = 'TCP'
            if str(port).upper() == 'ICMP':
                protocol = 'ICMP'
            if event_name == 'start':
                hookenv.open_port(port, protocol)
            elif event_name == 'stop':
                hookenv.close_port(port, protocol)

    def ports_contains(self, port, ports):
        if not bool(port):
            return False
        if str(port).upper() != 'ICMP':
            port = int(port)
        return port in ports


def service_stop(service_name):
    """
    Wrapper around host.service_stop to prevent spurious "unknown service"
    messages in the logs.
    """
    if host.service_running(service_name):
        host.service_stop(service_name)


def service_restart(service_name):
    """
    Wrapper around host.service_restart to prevent spurious "unknown service"
    messages in the logs.
    """
    if host.service_available(service_name):
        if host.service_running(service_name):
            host.service_restart(service_name)
        else:
            host.service_start(service_name)


# Convenience aliases
open_ports = close_ports = manage_ports = PortManagerCallback()



# Copyright 2016 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from charmhelpers.contrib.hardening.audits.file import TemplatedFile
from charmhelpers.contrib.hardening.host import TEMPLATES_DIR
from charmhelpers.contrib.hardening import utils


def get_audits():
    """Get OS hardening login.defs audits.

    :returns:  dictionary of audits
    """
    audits = [TemplatedFile('/etc/login.defs', LoginContext(),
                            template_dir=TEMPLATES_DIR,
                            user='root', group='root', mode=0o0444)]
    return audits


class LoginContext(object):

    def __call__(self):
        settings = utils.get_settings('os')

        # Octal numbers in yaml end up being turned into decimal,
        # so check if the umask is entered as a string (e.g. '027')
        # or as an octal umask as we know it (e.g. 002). If its not
        # a string assume it to be octal and turn it into an octal
        # string.
        umask = settings['environment']['umask']
        if not isinstance(umask, str):
            umask = '%s' % oct(umask)

        ctxt = {
            'additional_user_paths':
            settings['environment']['extra_user_paths'],
            'umask': umask,
            'pwd_max_age': settings['auth']['pw_max_age'],
            'pwd_min_age': settings['auth']['pw_min_age'],
            'uid_min': settings['auth']['uid_min'],
            'sys_uid_min': settings['auth']['sys_uid_min'],
            'sys_uid_max': settings['auth']['sys_uid_max'],
            'gid_min': settings['auth']['gid_min'],
            'sys_gid_min': settings['auth']['sys_gid_min'],
            'sys_gid_max': settings['auth']['sys_gid_max'],
            'login_retries': settings['auth']['retries'],
            'login_timeout': settings['auth']['timeout'],
            'chfn_restrict': settings['auth']['chfn_restrict'],
            'allow_login_without_home': settings['auth']['allow_homeless']
        }

        return ctxt



import json
import mock
import unittest

import charmhelpers.contrib.openstack.cert_utils as cert_utils


class CertUtilsTests(unittest.TestCase):

    def test_CertRequest(self):
        cr = cert_utils.CertRequest()
        self.assertEqual(cr.entries, [])
        self.assertIsNone(cr.hostname_entry)

    @mock.patch.object(cert_utils, 'local_unit', return_value='unit/2')
    def test_CertRequest_add_entry(self, local_unit):
        cr = cert_utils.CertRequest()
        cr.add_entry('admin', 'admin.openstack.local', ['10.10.10.10'])
        self.assertEqual(
            cr.get_request(),
            {'cert_requests':
                '{"admin.openstack.local": {"sans": ["10.10.10.10"]}}',
             'unit_name': 'unit_2'})

    @mock.patch.object(cert_utils, 'local_unit', return_value='unit/2')
    @mock.patch.object(cert_utils, 'resolve_network_cidr')
    @mock.patch.object(cert_utils, 'get_vip_in_network')
    @mock.patch.object(cert_utils, 'get_hostname')
    @mock.patch.object(cert_utils, 'local_address')
    def test_CertRequest_add_hostname_cn(self, local_address, get_hostname,
                                         get_vip_in_network,
                                         resolve_network_cidr, local_unit):
        resolve_network_cidr.side_effect = lambda x: x
        get_vip_in_network.return_value = '10.1.2.100'
        local_address.return_value = '10.1.2.3'
        get_hostname.return_value = 'juju-unit-2'
        cr = cert_utils.CertRequest()
        cr.add_hostname_cn()
        self.assertEqual(
            cr.get_request(),
            {'cert_requests':
                '{"juju-unit-2": {"sans": ["10.1.2.100", "10.1.2.3"]}}',
             'unit_name': 'unit_2'})

    @mock.patch.object(cert_utils, 'local_unit', return_value='unit/2')
    @mock.patch.object(cert_utils, 'resolve_network_cidr')
    @mock.patch.object(cert_utils, 'get_vip_in_network')
    @mock.patch.object(cert_utils, 'get_hostname')
    @mock.patch.object(cert_utils, 'local_address')
    def test_CertRequest_add_hostname_cn_ip(self, local_address, get_hostname,
                                            get_vip_in_network,
                                            resolve_network_cidr, local_unit):
        resolve_network_cidr.side_effect = lambda x: x
        get_vip_in_network.return_value = '10.1.2.100'
        local_address.return_value = '10.1.2.3'
        get_hostname.return_value = 'juju-unit-2'
        cr = cert_utils.CertRequest()
        cr.add_hostname_cn()
        cr.add_hostname_cn_ip(['10.1.2.4'])
        self.assertEqual(
            cr.get_request(),
            {'cert_requests':
                ('{"juju-unit-2": {"sans": ["10.1.2.100", "10.1.2.3", '
                 '"10.1.2.4"]}}'),
             'unit_name': 'unit_2'})

    @mock.patch.object(cert_utils, 'get_certificate_sans')
    @mock.patch.object(cert_utils, 'local_unit', return_value='unit/2')
    @mock.patch.object(cert_utils, 'resolve_network_cidr')
    @mock.patch.object(cert_utils, 'get_vip_in_network')
    @mock.patch.object(cert_utils, 'network_get_primary_address')
    @mock.patch.object(cert_utils, 'resolve_address')
    @mock.patch.object(cert_utils, 'config')
    @mock.patch.object(cert_utils, 'get_hostname')
    @mock.patch.object(cert_utils, 'local_address')
    def test_get_certificate_request(self, local_address, get_hostname,
                                     config, resolve_address,
                                     network_get_primary_address,
                                     get_vip_in_network, resolve_network_cidr,
                                     local_unit, get_certificate_sans):
        local_address.return_value = '10.1.2.3'
        get_hostname.return_value = 'juju-unit-2'
        _config = {
            'os-internal-hostname': 'internal.openstack.local',
            'os-admin-hostname': 'admin.openstack.local',
            'os-public-hostname': 'public.openstack.local',
        }
        _resolve_address = {
            'int': '10.0.0.2',
            'internal': '10.0.0.2',
            'admin': '10.10.0.2',
            'public': '10.20.0.2',
        }
        _npa = {
            'internal': '10.0.0.3',
            'admin': '10.10.0.3',
            'public': '10.20.0.3',
        }
        _vips = {
            '10.0.0.0/16': '10.0.0.100',
            '10.10.0.0/16': '10.10.0.100',
            '10.20.0.0/16': '10.20.0.100',
        }
        _resolve_nets = {
            '10.0.0.3': '10.0.0.0/16',
            '10.10.0.3': '10.10.0.0/16',
            '10.20.0.3': '10.20.0.0/16',
        }
        get_certificate_sans.return_value = list(set(
            list(_resolve_address.values()) +
            list(_npa.values()) +
            list(_vips.values())))
        expect = {
            'admin.openstack.local': {
                'sans': ['10.10.0.100', '10.10.0.2', '10.10.0.3']},
            'internal.openstack.local': {
                'sans': ['10.0.0.100', '10.0.0.2', '10.0.0.3']},
            'juju-unit-2': {'sans': ['10.1.2.3']},
            'public.openstack.local': {
                'sans': ['10.20.0.100', '10.20.0.2', '10.20.0.3']}}
        self.maxDiff = None
        config.side_effect = lambda x: _config.get(x)
        get_vip_in_network.side_effect = lambda x: _vips.get(x)
        resolve_network_cidr.side_effect = lambda x: _resolve_nets.get(x)
        network_get_primary_address.side_effect = lambda x: _npa.get(x)
        resolve_address.side_effect = \
            lambda endpoint_type: _resolve_address[endpoint_type]
        output = json.loads(
            cert_utils.get_certificate_request()['cert_requests'])
        self.assertEqual(
            output,
            expect)
        get_certificate_sans.assert_called_once_with(
            bindings=['internal', 'admin', 'public'])

    @mock.patch.object(cert_utils, 'get_certificate_sans')
    @mock.patch.object(cert_utils, 'local_unit', return_value='unit/2')
    @mock.patch.object(cert_utils, 'resolve_network_cidr')
    @mock.patch.object(cert_utils, 'get_vip_in_network')
    @mock.patch.object(cert_utils, 'network_get_primary_address')
    @mock.patch.object(cert_utils, 'resolve_address')
    @mock.patch.object(cert_utils, 'config')
    @mock.patch.object(cert_utils, 'get_hostname')
    @mock.patch.object(cert_utils, 'local_address')
    def test_get_certificate_request_no_hostnames(
            self, local_address, get_hostname, config, resolve_address,
            network_get_primary_address, get_vip_in_network,
            resolve_network_cidr, local_unit, get_certificate_sans):
        local_address.return_value = '10.1.2.3'
        get_hostname.return_value = 'juju-unit-2'
        _config = {
            'os-admin-hostname': 'admin.openstack.local',
            'os-public-hostname': 'public.openstack.local',
        }
        _resolve_address = {
            'int': '10.0.0.2',
            'internal': '10.0.0.2',
            'admin': '10.10.0.2',
            'public': '10.20.0.2',
        }
        _npa = {
            'internal': '10.0.0.3',
            'admin': '10.10.0.3',
            'public': '10.20.0.3',
            'mybinding': '10.30.0.3',
        }
        _vips = {
            '10.0.0.0/16': '10.0.0.100',
            '10.10.0.0/16': '10.10.0.100',
            '10.20.0.0/16': '10.20.0.100',
        }
        _resolve_nets = {
            '10.0.0.3': '10.0.0.0/16',
            '10.10.0.3': '10.10.0.0/16',
            '10.20.0.3': '10.20.0.0/16',
        }
        get_certificate_sans.return_value = list(set(
            list(_resolve_address.values()) +
            list(_npa.values()) +
            list(_vips.values())))
        expect = {
            'admin.openstack.local': {
                'sans': ['10.10.0.100', '10.10.0.2', '10.10.0.3']},
            'juju-unit-2': {'sans': [
                '10.0.0.100', '10.0.0.2', '10.0.0.3', '10.1.2.3', '10.30.0.3']},
            'public.openstack.local': {
                'sans': ['10.20.0.100', '10.20.0.2', '10.20.0.3']}}
        self.maxDiff = None
        config.side_effect = lambda x: _config.get(x)
        get_vip_in_network.side_effect = lambda x: _vips.get(x)
        resolve_network_cidr.side_effect = lambda x: _resolve_nets.get(x)
        network_get_primary_address.side_effect = lambda x: _npa.get(x)
        resolve_address.side_effect = \
            lambda endpoint_type: _resolve_address[endpoint_type]
        output = json.loads(
            cert_utils.get_certificate_request(
                bindings=['mybinding'])['cert_requests'])
        self.assertEqual(
            output,
            expect)
        get_certificate_sans.assert_called_once_with(
            bindings=['mybinding', 'internal', 'admin', 'public'])

    @mock.patch.object(cert_utils, 'get_certificate_request')
    @mock.patch.object(cert_utils, 'local_address')
    @mock.patch.object(cert_utils.os, 'symlink')
    @mock.patch.object(cert_utils.os.path, 'isfile')
    @mock.patch.object(cert_utils, 'get_hostname')
    def test_create_ip_cert_links(self, get_hostname, isfile,
                                  symlink, local_address, get_cert_request):
        cert_request = {'cert_requests': {
            'admin.openstack.local': {
                'sans': ['10.10.0.100', '10.10.0.2', '10.10.0.3']},
            'internal.openstack.local': {
                'sans': ['10.0.0.100', '10.0.0.2', '10.0.0.3']},
            'juju-unit-2': {'sans': ['10.1.2.3']},
            'public.openstack.local': {
                'sans': ['10.20.0.100', '10.20.0.2', '10.20.0.3']}}}
        get_cert_request.return_value = cert_request
        _files = {
            '/etc/ssl/cert_juju-unit-2': True,
            '/etc/ssl/cert_10.1.2.3': False,
            '/etc/ssl/cert_admin.openstack.local': True,
            '/etc/ssl/cert_10.10.0.100': False,
            '/etc/ssl/cert_10.10.0.2': False,
            '/etc/ssl/cert_10.10.0.3': False,
            '/etc/ssl/cert_internal.openstack.local': True,
            '/etc/ssl/cert_10.0.0.100': False,
            '/etc/ssl/cert_10.0.0.2': False,
            '/etc/ssl/cert_10.0.0.3': False,
            '/etc/ssl/cert_public.openstack.local': True,
            '/etc/ssl/cert_10.20.0.100': False,
            '/etc/ssl/cert_10.20.0.2': False,
            '/etc/ssl/cert_10.20.0.3': False,
            '/etc/ssl/cert_funky-name': False,
        }
        isfile.side_effect = lambda x: _files[x]
        expected = [
            mock.call('/etc/ssl/cert_admin.openstack.local', '/etc/ssl/cert_10.10.0.100'),
            mock.call('/etc/ssl/key_admin.openstack.local', '/etc/ssl/key_10.10.0.100'),
            mock.call('/etc/ssl/cert_admin.openstack.local', '/etc/ssl/cert_10.10.0.2'),
            mock.call('/etc/ssl/key_admin.openstack.local', '/etc/ssl/key_10.10.0.2'),
            mock.call('/etc/ssl/cert_admin.openstack.local', '/etc/ssl/cert_10.10.0.3'),
            mock.call('/etc/ssl/key_admin.openstack.local', '/etc/ssl/key_10.10.0.3'),
            mock.call('/etc/ssl/cert_internal.openstack.local', '/etc/ssl/cert_10.0.0.100'),
            mock.call('/etc/ssl/key_internal.openstack.local', '/etc/ssl/key_10.0.0.100'),
            mock.call('/etc/ssl/cert_internal.openstack.local', '/etc/ssl/cert_10.0.0.2'),
            mock.call('/etc/ssl/key_internal.openstack.local', '/etc/ssl/key_10.0.0.2'),
            mock.call('/etc/ssl/cert_internal.openstack.local', '/etc/ssl/cert_10.0.0.3'),
            mock.call('/etc/ssl/key_internal.openstack.local', '/etc/ssl/key_10.0.0.3'),
            mock.call('/etc/ssl/cert_juju-unit-2', '/etc/ssl/cert_10.1.2.3'),
            mock.call('/etc/ssl/key_juju-unit-2', '/etc/ssl/key_10.1.2.3'),
            mock.call('/etc/ssl/cert_public.openstack.local', '/etc/ssl/cert_10.20.0.100'),
            mock.call('/etc/ssl/key_public.openstack.local', '/etc/ssl/key_10.20.0.100'),
            mock.call('/etc/ssl/cert_public.openstack.local', '/etc/ssl/cert_10.20.0.2'),
            mock.call('/etc/ssl/key_public.openstack.local', '/etc/ssl/key_10.20.0.2'),
            mock.call('/etc/ssl/cert_public.openstack.local', '/etc/ssl/cert_10.20.0.3'),
            mock.call('/etc/ssl/key_public.openstack.local', '/etc/ssl/key_10.20.0.3')]
        cert_utils.create_ip_cert_links('/etc/ssl')
        symlink.assert_has_calls(expected, any_order=True)
        # Customer hostname
        symlink.reset_mock()
        get_hostname.return_value = 'juju-unit-2'
        cert_utils.create_ip_cert_links(
            '/etc/ssl',
            custom_hostname_link='funky-name')
        expected.extend([
            mock.call('/etc/ssl/cert_juju-unit-2', '/etc/ssl/cert_funky-name'),
            mock.call('/etc/ssl/key_juju-unit-2', '/etc/ssl/key_funky-name'),
        ])
        symlink.assert_has_calls(expected, any_order=True)
        get_cert_request.assert_called_with(
            json_encode=False, bindings=['internal', 'admin', 'public'])

    @mock.patch.object(cert_utils, 'get_certificate_request')
    @mock.patch.object(cert_utils, 'local_address')
    @mock.patch.object(cert_utils.os, 'symlink')
    @mock.patch.object(cert_utils.os.path, 'isfile')
    @mock.patch.object(cert_utils, 'get_hostname')
    def test_create_ip_cert_links_bindings(
            self, get_hostname, isfile, symlink, local_address, get_cert_request):
        cert_request = {'cert_requests': {
            'admin.openstack.local': {
                'sans': ['10.10.0.100', '10.10.0.2', '10.10.0.3']},
            'internal.openstack.local': {
                'sans': ['10.0.0.100', '10.0.0.2', '10.0.0.3']},
            'juju-unit-2': {'sans': ['10.1.2.3']},
            'public.openstack.local': {
                'sans': ['10.20.0.100', '10.20.0.2', '10.20.0.3']}}}
        get_cert_request.return_value = cert_request
        _files = {
            '/etc/ssl/cert_juju-unit-2': True,
            '/etc/ssl/cert_10.1.2.3': False,
            '/etc/ssl/cert_admin.openstack.local': True,
            '/etc/ssl/cert_10.10.0.100': False,
            '/etc/ssl/cert_10.10.0.2': False,
            '/etc/ssl/cert_10.10.0.3': False,
            '/etc/ssl/cert_internal.openstack.local': True,
            '/etc/ssl/cert_10.0.0.100': False,
            '/etc/ssl/cert_10.0.0.2': False,
            '/etc/ssl/cert_10.0.0.3': False,
            '/etc/ssl/cert_public.openstack.local': True,
            '/etc/ssl/cert_10.20.0.100': False,
            '/etc/ssl/cert_10.20.0.2': False,
            '/etc/ssl/cert_10.20.0.3': False,
            '/etc/ssl/cert_funky-name': False,
        }
        isfile.side_effect = lambda x: _files[x]
        expected = [
            mock.call('/etc/ssl/cert_admin.openstack.local', '/etc/ssl/cert_10.10.0.100'),
            mock.call('/etc/ssl/key_admin.openstack.local', '/etc/ssl/key_10.10.0.100'),
            mock.call('/etc/ssl/cert_admin.openstack.local', '/etc/ssl/cert_10.10.0.2'),
            mock.call('/etc/ssl/key_admin.openstack.local', '/etc/ssl/key_10.10.0.2'),
            mock.call('/etc/ssl/cert_admin.openstack.local', '/etc/ssl/cert_10.10.0.3'),
            mock.call('/etc/ssl/key_admin.openstack.local', '/etc/ssl/key_10.10.0.3'),
            mock.call('/etc/ssl/cert_internal.openstack.local', '/etc/ssl/cert_10.0.0.100'),
            mock.call('/etc/ssl/key_internal.openstack.local', '/etc/ssl/key_10.0.0.100'),
            mock.call('/etc/ssl/cert_internal.openstack.local', '/etc/ssl/cert_10.0.0.2'),
            mock.call('/etc/ssl/key_internal.openstack.local', '/etc/ssl/key_10.0.0.2'),
            mock.call('/etc/ssl/cert_internal.openstack.local', '/etc/ssl/cert_10.0.0.3'),
            mock.call('/etc/ssl/key_internal.openstack.local', '/etc/ssl/key_10.0.0.3'),
            mock.call('/etc/ssl/cert_juju-unit-2', '/etc/ssl/cert_10.1.2.3'),
            mock.call('/etc/ssl/key_juju-unit-2', '/etc/ssl/key_10.1.2.3'),
            mock.call('/etc/ssl/cert_public.openstack.local', '/etc/ssl/cert_10.20.0.100'),
            mock.call('/etc/ssl/key_public.openstack.local', '/etc/ssl/key_10.20.0.100'),
            mock.call('/etc/ssl/cert_public.openstack.local', '/etc/ssl/cert_10.20.0.2'),
            mock.call('/etc/ssl/key_public.openstack.local', '/etc/ssl/key_10.20.0.2'),
            mock.call('/etc/ssl/cert_public.openstack.local', '/etc/ssl/cert_10.20.0.3'),
            mock.call('/etc/ssl/key_public.openstack.local', '/etc/ssl/key_10.20.0.3')]
        cert_utils.create_ip_cert_links('/etc/ssl', bindings=['mybindings'])
        symlink.assert_has_calls(expected, any_order=True)
        # Customer hostname
        symlink.reset_mock()
        get_hostname.return_value = 'juju-unit-2'
        cert_utils.create_ip_cert_links(
            '/etc/ssl',
            custom_hostname_link='funky-name', bindings=['mybinding'])
        expected.extend([
            mock.call('/etc/ssl/cert_juju-unit-2', '/etc/ssl/cert_funky-name'),
            mock.call('/etc/ssl/key_juju-unit-2', '/etc/ssl/key_funky-name'),
        ])
        symlink.assert_has_calls(expected, any_order=True)
        get_cert_request.assert_called_with(
            json_encode=False, bindings=['mybinding', 'internal', 'admin', 'public'])

    @mock.patch.object(cert_utils, 'write_file')
    def test_install_certs(self, write_file):
        certs = {
            'admin.openstack.local': {
                'cert': 'ADMINCERT',
                'key': 'ADMINKEY'}}
        cert_utils.install_certs('/etc/ssl', certs, chain='CHAIN')
        expected = [
            mock.call(
                path='/etc/ssl/cert_admin.openstack.local',
                content='ADMINCERT\nCHAIN',
                owner='root', group='root',
                perms=0o640),
            mock.call(
                path='/etc/ssl/key_admin.openstack.local',
                content='ADMINKEY',
                owner='root', group='root',
                perms=0o640),
        ]
        write_file.assert_has_calls(expected)

    @mock.patch.object(cert_utils, 'write_file')
    def test_install_certs_ca(self, write_file):
        certs = {
            'admin.openstack.local': {
                'cert': 'ADMINCERT',
                'key': 'ADMINKEY'}}
        ca = 'MYCA'
        cert_utils.install_certs('/etc/ssl', certs, ca)
        expected = [
            mock.call(
                path='/etc/ssl/cert_admin.openstack.local',
                content='ADMINCERT\nMYCA',
                owner='root', group='root',
                perms=0o640),
            mock.call(
                path='/etc/ssl/key_admin.openstack.local',
                content='ADMINKEY',
                owner='root', group='root',
                perms=0o640),
        ]
        write_file.assert_has_calls(expected)

    @mock.patch.object(cert_utils, '_manage_ca_certs')
    @mock.patch.object(cert_utils, 'remote_service_name')
    @mock.patch.object(cert_utils, 'local_unit')
    @mock.patch.object(cert_utils, 'create_ip_cert_links')
    @mock.patch.object(cert_utils, 'install_certs')
    @mock.patch.object(cert_utils, 'install_ca_cert')
    @mock.patch.object(cert_utils, 'mkdir')
    @mock.patch.object(cert_utils, 'relation_get')
    def test_process_certificates(self, relation_get, mkdir, install_ca_cert,
                                  install_certs, create_ip_cert_links,
                                  local_unit, remote_service_name,
                                  _manage_ca_certs):
        remote_service_name.return_value = 'vault'
        local_unit.return_value = 'devnull/2'
        certs = {
            'admin.openstack.local': {
                'cert': 'ADMINCERT',
                'key': 'ADMINKEY'}}
        _relation_info = {
            'keystone_2.processed_requests': json.dumps(certs),
            'chain': 'MYCHAIN',
            'ca': 'ROOTCA',
        }
        relation_get.return_value = _relation_info
        self.assertFalse(cert_utils.process_certificates(
            'myservice',
            'certificates:2',
            'vault/0',
            custom_hostname_link='funky-name'))
        local_unit.return_value = 'keystone/2'
        self.assertTrue(cert_utils.process_certificates(
            'myservice',
            'certificates:2',
            'vault/0',
            custom_hostname_link='funky-name'))
        _manage_ca_certs.assert_called_once_with(
            'ROOTCA', 'certificates:2')
        install_certs.assert_called_once_with(
            '/etc/apache2/ssl/myservice',
            {'admin.openstack.local': {
                'key': 'ADMINKEY', 'cert': 'ADMINCERT'}},
            'MYCHAIN', user='root', group='root')
        create_ip_cert_links.assert_called_once_with(
            '/etc/apache2/ssl/myservice',
            custom_hostname_link='funky-name',
            bindings=['internal', 'admin', 'public'])

    @mock.patch.object(cert_utils, '_manage_ca_certs')
    @mock.patch.object(cert_utils, 'remote_service_name')
    @mock.patch.object(cert_utils, 'local_unit')
    @mock.patch.object(cert_utils, 'create_ip_cert_links')
    @mock.patch.object(cert_utils, 'install_certs')
    @mock.patch.object(cert_utils, 'install_ca_cert')
    @mock.patch.object(cert_utils, 'mkdir')
    @mock.patch.object(cert_utils, 'relation_get')
    def test_process_certificates_bindings(
            self, relation_get, mkdir, install_ca_cert,
            install_certs, create_ip_cert_links,
            local_unit, remote_service_name, _manage_ca_certs):
        remote_service_name.return_value = 'vault'
        local_unit.return_value = 'devnull/2'
        certs = {
            'admin.openstack.local': {
                'cert': 'ADMINCERT',
                'key': 'ADMINKEY'}}
        _relation_info = {
            'keystone_2.processed_requests': json.dumps(certs),
            'chain': 'MYCHAIN',
            'ca': 'ROOTCA',
        }
        relation_get.return_value = _relation_info
        self.assertFalse(cert_utils.process_certificates(
            'myservice',
            'certificates:2',
            'vault/0',
            custom_hostname_link='funky-name'))
        local_unit.return_value = 'keystone/2'
        self.assertTrue(cert_utils.process_certificates(
            'myservice',
            'certificates:2',
            'vault/0',
            custom_hostname_link='funky-name',
            bindings=['mybinding']))
        _manage_ca_certs.assert_called_once_with(
            'ROOTCA', 'certificates:2')
        install_certs.assert_called_once_with(
            '/etc/apache2/ssl/myservice',
            {'admin.openstack.local': {
                'key': 'ADMINKEY', 'cert': 'ADMINCERT'}},
            'MYCHAIN', user='root', group='root')
        create_ip_cert_links.assert_called_once_with(
            '/etc/apache2/ssl/myservice',
            custom_hostname_link='funky-name',
            bindings=['mybinding', 'internal', 'admin', 'public'])

    @mock.patch.object(cert_utils, 'remote_service_name')
    @mock.patch.object(cert_utils, 'relation_ids')
    def test_get_cert_relation_ca_name(self, relation_ids, remote_service_name):
        remote_service_name.return_value = 'vault'

        # Test with argument:
        self.assertEqual(cert_utils.get_cert_relation_ca_name('certificates:1'),
                         'vault_juju_ca_cert')
        remote_service_name.assert_called_once_with(relid='certificates:1')
        remote_service_name.reset_mock()

        # Test without argument:
        relation_ids.return_value = ['certificates:2']
        self.assertEqual(cert_utils.get_cert_relation_ca_name(),
                         'vault_juju_ca_cert')
        remote_service_name.assert_called_once_with(relid='certificates:2')
        remote_service_name.reset_mock()

        # Test without argument nor 'certificates' relation:
        relation_ids.return_value = []
        self.assertEqual(cert_utils.get_cert_relation_ca_name(), '')
        remote_service_name.assert_not_called()

    @mock.patch.object(cert_utils, 'log')
    @mock.patch.object(cert_utils, 'remote_service_name')
    @mock.patch.object(cert_utils.os, 'remove')
    @mock.patch.object(cert_utils.os.path, 'exists')
    @mock.patch.object(cert_utils, 'config')
    @mock.patch.object(cert_utils, 'install_ca_cert')
    def test__manage_ca_certs(self, install_ca_cert, config, os_exists,
                              os_remove, remote_service_name, _log):
        remote_service_name.return_value = 'vault'
        _config = {}
        config.side_effect = lambda x: _config.get(x)
        os_exists.return_value = False
        cert_utils._manage_ca_certs('CA', 'certificates:2')
        install_ca_cert.assert_called_once_with(
            b'CA',
            name='vault_juju_ca_cert')
        self.assertFalse(os_remove.called)
        # Test old cert removed.
        install_ca_cert.reset_mock()
        os_exists.reset_mock()
        os_exists.return_value = True
        cert_utils._manage_ca_certs('CA', 'certificates:2')
        install_ca_cert.assert_called_once_with(
            b'CA',
            name='vault_juju_ca_cert')
        os_remove.assert_called_once_with(
            '/usr/local/share/ca-certificates/keystone_juju_ca_cert.crt')
        # Test cert is installed from config
        _config['ssl_ca'] = 'Q0FGUk9NQ09ORklHCg=='
        install_ca_cert.reset_mock()
        os_remove.reset_mock()
        os_exists.reset_mock()
        os_exists.return_value = True
        cert_utils._manage_ca_certs('CA', 'certificates:2')
        expected = [
            mock.call(b'CAFROMCONFIG', name='keystone_juju_ca_cert'),
            mock.call(b'CA', name='vault_juju_ca_cert')]
        install_ca_cert.assert_has_calls(expected)
        self.assertFalse(os_remove.called)

    @mock.patch.object(cert_utils, 'local_unit')
    @mock.patch.object(cert_utils, 'related_units')
    @mock.patch.object(cert_utils, 'relation_ids')
    @mock.patch.object(cert_utils, 'relation_get')
    def test_get_requests_for_local_unit(self, relation_get, relation_ids,
                                         related_units, local_unit):
        local_unit.return_value = 'rabbitmq-server/2'
        relation_ids.return_value = ['certificates:12']
        related_units.return_value = ['vault/0']
        certs = {
            'juju-cd4bb3-5.lxd': {
                'cert': 'BASECERT',
                'key': 'BASEKEY'},
            'juju-cd4bb3-5.internal': {
                'cert': 'INTERNALCERT',
                'key': 'INTERNALKEY'}}
        _relation_info = {
            'rabbitmq-server_2.processed_requests': json.dumps(certs),
            'chain': 'MYCHAIN',
            'ca': 'ROOTCA',
        }
        relation_get.return_value = _relation_info
        self.assertEqual(
            cert_utils.get_requests_for_local_unit(),
            [{
                'ca': 'ROOTCA',
                'certs': {
                    'juju-cd4bb3-5.lxd': {
                        'cert': 'BASECERT',
                        'key': 'BASEKEY'},
                    'juju-cd4bb3-5.internal': {
                        'cert': 'INTERNALCERT',
                        'key': 'INTERNALKEY'}},
                'chain': 'MYCHAIN'}]
        )

    @mock.patch.object(cert_utils, 'get_requests_for_local_unit')
    def test_get_bundle_for_cn(self, get_requests_for_local_unit):
        get_requests_for_local_unit.return_value = [{
            'ca': 'ROOTCA',
            'certs': {
                'juju-cd4bb3-5.lxd': {
                    'cert': 'BASECERT',
                    'key': 'BASEKEY'},
                'juju-cd4bb3-5.internal': {
                    'cert': 'INTERNALCERT',
                    'key': 'INTERNALKEY'}},
            'chain': 'MYCHAIN'}]
        self.assertEqual(
            cert_utils.get_bundle_for_cn('juju-cd4bb3-5.internal'),
            {
                'ca': 'ROOTCA',
                'cert': 'INTERNALCERT',
                'chain': 'MYCHAIN',
                'key': 'INTERNALKEY'})

    @mock.patch.object(cert_utils, 'local_unit', return_value='unit/2')
    @mock.patch.object(cert_utils, 'resolve_network_cidr')
    @mock.patch.object(cert_utils, 'get_vip_in_network')
    @mock.patch.object(cert_utils, 'get_relation_ip')
    @mock.patch.object(cert_utils, 'resolve_address')
    @mock.patch.object(cert_utils, 'config')
    @mock.patch.object(cert_utils, 'get_hostname')
    @mock.patch.object(cert_utils, 'local_address')
    def test_get_certificate_sans(self, local_address, get_hostname,
                                  config, resolve_address,
                                  get_relation_ip,
                                  get_vip_in_network, resolve_network_cidr,
                                  local_unit):
        local_address.return_value = '10.1.2.3'
        get_hostname.return_value = 'juju-unit-2'
        _config = {
            'os-internal-hostname': 'internal.openstack.local',
            'os-admin-hostname': 'admin.openstack.local',
            'os-public-hostname': 'public.openstack.local',
        }
        _resolve_address = {
            'int': '10.0.0.2',
            'internal': '10.0.0.2',
            'admin': '10.10.0.2',
            'public': '10.20.0.2',
        }
        _npa = {
            'internal': '10.0.0.3',
            'admin': '10.10.0.3',
            'public': '10.20.0.3',
        }
        _vips = {
            '10.0.0.0/16': '10.0.0.100',
            '10.10.0.0/16': '10.10.0.100',
            '10.20.0.0/16': '10.20.0.100',
        }
        _resolve_nets = {
            '10.0.0.3': '10.0.0.0/16',
            '10.10.0.3': '10.10.0.0/16',
            '10.20.0.3': '10.20.0.0/16',
        }
        expect = list(set([
            '10.10.0.100', '10.10.0.2', '10.10.0.3',
            '10.0.0.100', '10.0.0.2', '10.0.0.3',
            '10.1.2.3',
            '10.20.0.100', '10.20.0.2', '10.20.0.3']))
        self.maxDiff = None
        config.side_effect = lambda x: _config.get(x)
        get_vip_in_network.side_effect = lambda x: _vips.get(x)
        resolve_network_cidr.side_effect = lambda x: _resolve_nets.get(x)
        get_relation_ip.side_effect = lambda x, cidr_network: _npa.get(x)
        resolve_address.side_effect = \
            lambda endpoint_type: _resolve_address[endpoint_type]
        expected_get_relation_ip_calls = [
            mock.call('internal', cidr_network=None),
            mock.call('admin', cidr_network=None),
            mock.call('public', cidr_network=None)]
        self.assertEqual(cert_utils.get_certificate_sans().sort(),
                         expect.sort())
        get_relation_ip.assert_has_calls(
            expected_get_relation_ip_calls, any_order=True)

    @mock.patch.object(cert_utils, 'local_unit', return_value='unit/2')
    @mock.patch.object(cert_utils, 'resolve_network_cidr')
    @mock.patch.object(cert_utils, 'get_vip_in_network')
    @mock.patch.object(cert_utils, 'get_relation_ip')
    @mock.patch.object(cert_utils, 'resolve_address')
    @mock.patch.object(cert_utils, 'config')
    @mock.patch.object(cert_utils, 'get_hostname')
    @mock.patch.object(cert_utils, 'local_address')
    def test_get_certificate_sans_bindings(
            self, local_address, get_hostname, config, resolve_address,
            get_relation_ip, get_vip_in_network, resolve_network_cidr, local_unit):
        local_address.return_value = '10.1.2.3'
        get_hostname.return_value = 'juju-unit-2'
        _config = {
            'os-internal-hostname': 'internal.openstack.local',
            'os-admin-hostname': 'admin.openstack.local',
            'os-public-hostname': 'public.openstack.local',
        }
        _resolve_address = {
            'int': '10.0.0.2',
            'internal': '10.0.0.2',
            'admin': '10.10.0.2',
            'public': '10.20.0.2',
        }
        _npa = {
            'internal': '10.0.0.3',
            'admin': '10.10.0.3',
            'public': '10.20.0.3',
        }
        _vips = {
            '10.0.0.0/16': '10.0.0.100',
            '10.10.0.0/16': '10.10.0.100',
            '10.20.0.0/16': '10.20.0.100',
        }
        _resolve_nets = {
            '10.0.0.3': '10.0.0.0/16',
            '10.10.0.3': '10.10.0.0/16',
            '10.20.0.3': '10.20.0.0/16',
        }
        expect = list(set([
            '10.10.0.100', '10.10.0.2', '10.10.0.3',
            '10.0.0.100', '10.0.0.2', '10.0.0.3',
            '10.1.2.3',
            '10.20.0.100', '10.20.0.2', '10.20.0.3']))
        self.maxDiff = None
        config.side_effect = lambda x: _config.get(x)
        get_vip_in_network.side_effect = lambda x: _vips.get(x)
        resolve_network_cidr.side_effect = lambda x: _resolve_nets.get(x)
        get_relation_ip.side_effect = lambda x, cidr_network: _npa.get(x)
        resolve_address.side_effect = \
            lambda endpoint_type: _resolve_address[endpoint_type]
        expected_get_relation_ip_calls = [
            mock.call('internal', cidr_network=None),
            mock.call('admin', cidr_network=None),
            mock.call('public', cidr_network=None),
            mock.call('mybinding', cidr_network=None)]
        self.assertEqual(
            cert_utils.get_certificate_sans(bindings=['mybinding']).sort(),
            expect.sort())
        get_relation_ip.assert_has_calls(
            expected_get_relation_ip_calls, any_order=True)



from mock import patch, MagicMock, call

from subprocess import CalledProcessError
from testtools import TestCase

import charmhelpers.contrib.hahelpers.cluster as cluster_utils

CRM_STATUS = b'''
Last updated: Thu May 14 14:46:35 2015
Last change: Thu May 14 14:43:51 2015 via crmd on juju-trusty-machine-1
Stack: corosync
Current DC: juju-trusty-machine-2 (168108171) - partition with quorum
Version: 1.1.10-42f2063
3 Nodes configured
4 Resources configured


Online: [ juju-trusty-machine-1 juju-trusty-machine-2 juju-trusty-machine-3 ]

 Resource Group: grp_percona_cluster
      res_mysql_vip      (ocf::heartbeat:IPaddr2):       Started juju-trusty-machine-1
       Clone Set: cl_mysql_monitor [res_mysql_monitor]
            Started: [ juju-trusty-machine-1 juju-trusty-machine-2 juju-trusty-machine-3 ]
'''

CRM_DC_NONE = b'''
Last updated: Thu May 14 14:46:35 2015
Last change: Thu May 14 14:43:51 2015 via crmd on juju-trusty-machine-1
Stack: corosync
Current DC: NONE
1 Nodes configured, 2 expected votes
0 Resources configured


Node node1: UNCLEAN (offline)
'''


class ClusterUtilsTests(TestCase):
    def setUp(self):
        super(ClusterUtilsTests, self).setUp()
        [self._patch(m) for m in [
            'log',
            'relation_ids',
            'relation_list',
            'relation_get',
            'get_unit_hostname',
            'config_get',
            'unit_get',
        ]]

    def _patch(self, method):
        _m = patch.object(cluster_utils, method)
        mock = _m.start()
        self.addCleanup(_m.stop)
        setattr(self, method, mock)

    def test_is_clustered(self):
        '''It determines whether or not a unit is clustered'''
        self.relation_ids.return_value = ['ha:0']
        self.relation_list.return_value = ['ha/0']
        self.relation_get.return_value = 'yes'
        self.assertTrue(cluster_utils.is_clustered())

    def test_is_not_clustered(self):
        '''It determines whether or not a unit is clustered'''
        self.relation_ids.return_value = ['ha:0']
        self.relation_list.return_value = ['ha/0']
        self.relation_get.return_value = None
        self.assertFalse(cluster_utils.is_clustered())

    @patch('subprocess.check_output')
    def test_is_crm_dc(self, check_output):
        '''It determines its unit is leader'''
        self.get_unit_hostname.return_value = 'juju-trusty-machine-2'
        check_output.return_value = CRM_STATUS
        self.assertTrue(cluster_utils.is_crm_dc())

    @patch('subprocess.check_output')
    def test_is_crm_dc_no_cluster(self, check_output):
        '''It is not leader if there is no cluster up'''
        def r(*args, **kwargs):
            raise CalledProcessError(1, 'crm')

        check_output.side_effect = r
        self.assertRaises(cluster_utils.CRMDCNotFound, cluster_utils.is_crm_dc)

    @patch('subprocess.check_output')
    def test_is_crm_dc_false(self, check_output):
        '''It determines its unit is leader'''
        self.get_unit_hostname.return_value = 'juju-trusty-machine-1'
        check_output.return_value = CRM_STATUS
        self.assertFalse(cluster_utils.is_crm_dc())

    @patch('subprocess.check_output')
    def test_is_crm_dc_current_none(self, check_output):
        '''It determines its unit is leader'''
        self.get_unit_hostname.return_value = 'juju-trusty-machine-1'
        check_output.return_value = CRM_DC_NONE
        self.assertRaises(cluster_utils.CRMDCNotFound, cluster_utils.is_crm_dc)

    @patch('subprocess.check_output')
    def test_is_crm_leader(self, check_output):
        '''It determines its unit is leader'''
        self.get_unit_hostname.return_value = 'node1'
        crm = b'resource vip is running on: node1'
        check_output.return_value = crm
        self.assertTrue(cluster_utils.is_crm_leader('vip'))

    @patch('charmhelpers.core.decorators.time')
    @patch('subprocess.check_output')
    def test_is_not_leader(self, check_output, mock_time):
        '''It determines its unit is not leader'''
        self.get_unit_hostname.return_value = 'node1'
        crm = b'resource vip is running on: node2'
        check_output.return_value = crm
        self.assertFalse(cluster_utils.is_crm_leader('some_resource'))
        self.assertFalse(mock_time.called)

    @patch('charmhelpers.core.decorators.log')
    @patch('charmhelpers.core.decorators.time')
    @patch('subprocess.check_output')
    def test_is_not_leader_resource_not_exists(self, check_output, mock_time,
                                               mock_log):
        '''It determines its unit is not leader'''
        self.get_unit_hostname.return_value = 'node1'
        check_output.return_value = b"resource vip is NOT running"
        self.assertRaises(cluster_utils.CRMResourceNotFound,
                          cluster_utils.is_crm_leader, 'vip')
        mock_time.assert_has_calls([call.sleep(2), call.sleep(4),
                                    call.sleep(6)])

    @patch('charmhelpers.core.decorators.time')
    @patch('subprocess.check_output')
    def test_is_crm_leader_no_cluster(self, check_output, mock_time):
        '''It is not leader if there is no cluster up'''
        check_output.side_effect = CalledProcessError(1, 'crm')
        self.assertFalse(cluster_utils.is_crm_leader('vip'))
        self.assertFalse(mock_time.called)

    @patch.object(cluster_utils, 'is_crm_dc')
    def test_is_crm_leader_dc_resource(self, _is_crm_dc):
        '''Call out to is_crm_dc'''
        cluster_utils.is_crm_leader(cluster_utils.DC_RESOURCE_NAME)
        _is_crm_dc.assert_called_with()

    def test_peer_units(self):
        '''It lists all peer units for cluster relation'''
        peers = ['peer_node/1', 'peer_node/2']
        self.relation_ids.return_value = ['cluster:0']
        self.relation_list.return_value = peers
        self.assertEquals(peers, cluster_utils.peer_units())

    def test_peer_ips(self):
        '''Get a dict of peers and their ips'''
        peers = {
            'peer_node/1': '10.0.0.1',
            'peer_node/2': '10.0.0.2',
        }

        def _relation_get(attr, rid, unit):
            return peers[unit]
        self.relation_ids.return_value = ['cluster:0']
        self.relation_list.return_value = peers.keys()
        self.relation_get.side_effect = _relation_get
        self.assertEquals(peers, cluster_utils.peer_ips())

    @patch('os.getenv')
    def test_is_oldest_peer(self, getenv):
        '''It detects local unit is the oldest of all peers'''
        peers = ['peer_node/1', 'peer_node/2', 'peer_node/3']
        getenv.return_value = 'peer_node/1'
        self.assertTrue(cluster_utils.oldest_peer(peers))

    @patch('os.getenv')
    def test_is_not_oldest_peer(self, getenv):
        '''It detects local unit is not the oldest of all peers'''
        peers = ['peer_node/1', 'peer_node/2', 'peer_node/3']
        getenv.return_value = 'peer_node/2'
        self.assertFalse(cluster_utils.oldest_peer(peers))

    @patch.object(cluster_utils, 'is_crm_leader')
    @patch.object(cluster_utils, 'is_clustered')
    def test_is_elected_leader_clustered(self, is_clustered, is_crm_leader):
        '''It detects it is the eligible leader in a hacluster of units'''
        is_clustered.return_value = True
        is_crm_leader.return_value = True
        self.assertTrue(cluster_utils.is_elected_leader('vip'))

    @patch.object(cluster_utils, 'is_crm_leader')
    @patch.object(cluster_utils, 'is_clustered')
    def test_not_is_elected_leader_clustered(self, is_clustered, is_crm_leader):
        '''It detects it is not the eligible leader in a hacluster of units'''
        is_clustered.return_value = True
        is_crm_leader.return_value = False
        self.assertFalse(cluster_utils.is_elected_leader('vip'))

    @patch.object(cluster_utils, 'oldest_peer')
    @patch.object(cluster_utils, 'peer_units')
    @patch.object(cluster_utils, 'is_clustered')
    def test_is_is_elected_leader_unclustered(self, is_clustered,
                                              peer_units, oldest_peer):
        '''It detects it is the eligible leader in non-clustered peer group'''
        is_clustered.return_value = False
        oldest_peer.return_value = True
        self.assertTrue(cluster_utils.is_elected_leader('vip'))

    @patch.object(cluster_utils, 'oldest_peer')
    @patch.object(cluster_utils, 'peer_units')
    @patch.object(cluster_utils, 'is_clustered')
    def test_not_is_elected_leader_unclustered(self, is_clustered,
                                               peer_units, oldest_peer):
        '''It detects it is not the eligible leader in non-clustered group'''
        is_clustered.return_value = False
        oldest_peer.return_value = False
        self.assertFalse(cluster_utils.is_elected_leader('vip'))

    def test_https_explict(self):
        '''It determines https is available if configured explicitly'''
        # config_get('use-https')
        self.config_get.return_value = 'yes'
        self.assertTrue(cluster_utils.https())

    def test_https_cert_key_in_config(self):
        '''It determines https is available if cert + key in charm config'''
        # config_get('use-https')
        self.config_get.side_effect = [
            'no',  # config_get('use-https')
            'cert',  # config_get('ssl_cert')
            'key',  # config_get('ssl_key')
        ]
        self.assertTrue(cluster_utils.https())

    def test_https_cert_key_in_identity_relation(self):
        '''It determines https is available if cert in identity-service'''
        self.config_get.return_value = False
        self.relation_ids.return_value = 'identity-service:0'
        self.relation_list.return_value = 'keystone/0'
        self.relation_get.side_effect = [
            'yes',  # relation_get('https_keystone')
            'cert',  # relation_get('ssl_cert')
            'key',  # relation_get('ssl_key')
            'ca_cert',  # relation_get('ca_cert')
        ]
        self.assertTrue(cluster_utils.https())

    def test_https_cert_key_incomplete_identity_relation(self):
        '''It determines https unavailable if cert not in identity-service'''
        self.config_get.return_value = False
        self.relation_ids.return_value = 'identity-service:0'
        self.relation_list.return_value = 'keystone/0'
        self.relation_get.return_value = None
        self.assertFalse(cluster_utils.https())

    @patch.object(cluster_utils, 'https')
    @patch.object(cluster_utils, 'peer_units')
    def test_determine_api_port_with_peers(self, peer_units, https):
        '''It determines API port in presence of peers'''
        peer_units.return_value = ['peer1']
        https.return_value = False
        self.assertEquals(9686, cluster_utils.determine_api_port(9696))

    @patch.object(cluster_utils, 'https')
    @patch.object(cluster_utils, 'peer_units')
    def test_determine_api_port_nopeers_singlemode(self, peer_units, https):
        '''It determines API port with a single unit in singlemode'''
        peer_units.return_value = []
        https.return_value = False
        port = cluster_utils.determine_api_port(9696, singlenode_mode=True)
        self.assertEquals(9686, port)

    @patch.object(cluster_utils, 'is_clustered')
    @patch.object(cluster_utils, 'https')
    @patch.object(cluster_utils, 'peer_units')
    def test_determine_api_port_clustered(self, peer_units, https,
                                          is_clustered):
        '''It determines API port in presence of an hacluster'''
        peer_units.return_value = []
        is_clustered.return_value = True
        https.return_value = False
        self.assertEquals(9686, cluster_utils.determine_api_port(9696))

    @patch.object(cluster_utils, 'is_clustered')
    @patch.object(cluster_utils, 'https')
    @patch.object(cluster_utils, 'peer_units')
    def test_determine_api_port_clustered_https(self, peer_units, https,
                                                is_clustered):
        '''It determines API port in presence of hacluster + https'''
        peer_units.return_value = []
        is_clustered.return_value = True
        https.return_value = True
        self.assertEquals(9676, cluster_utils.determine_api_port(9696))

    @patch.object(cluster_utils, 'https')
    def test_determine_apache_port_https(self, https):
        '''It determines haproxy port with https enabled'''
        https.return_value = True
        self.assertEquals(9696, cluster_utils.determine_apache_port(9696))

    @patch.object(cluster_utils, 'https')
    @patch.object(cluster_utils, 'is_clustered')
    def test_determine_apache_port_clustered(self, https, is_clustered):
        '''It determines haproxy port with https disabled'''
        https.return_value = True
        is_clustered.return_value = True
        self.assertEquals(9686, cluster_utils.determine_apache_port(9696))

    @patch.object(cluster_utils, 'peer_units')
    @patch.object(cluster_utils, 'https')
    @patch.object(cluster_utils, 'is_clustered')
    def test_determine_apache_port_nopeers_singlemode(self, https,
                                                      is_clustered,
                                                      peer_units):
        '''It determines haproxy port with a single unit in singlemode'''
        peer_units.return_value = []
        https.return_value = False
        is_clustered.return_value = False
        port = cluster_utils.determine_apache_port(9696, singlenode_mode=True)
        self.assertEquals(9686, port)

    @patch.object(cluster_utils, 'valid_hacluster_config')
    def test_get_hacluster_config_complete(self, valid_hacluster_config):
        '''It fetches all hacluster charm config'''
        conf = {
            'ha-bindiface': 'eth1',
            'ha-mcastport': '3333',
            'vip': '10.0.0.1',
            'os-admin-hostname': None,
            'os-public-hostname': None,
            'os-internal-hostname': None,
            'os-access-hostname': None,
        }

        valid_hacluster_config.return_value = True

        def _fake_config_get(setting):
            return conf[setting]

        self.config_get.side_effect = _fake_config_get
        self.assertEquals(conf, cluster_utils.get_hacluster_config())

    @patch.object(cluster_utils, 'valid_hacluster_config')
    def test_get_hacluster_config_incomplete(self, valid_hacluster_config):
        '''It raises exception if some hacluster charm config missing'''
        conf = {
            'ha-bindiface': 'eth1',
            'ha-mcastport': '3333',
            'vip': None,
            'os-admin-hostname': None,
            'os-public-hostname': None,
            'os-internal-hostname': None,
            'os-access-hostname': None,
        }

        valid_hacluster_config.return_value = False

        def _fake_config_get(setting):
            return conf[setting]

        self.config_get.side_effect = _fake_config_get
        self.assertRaises(cluster_utils.HAIncorrectConfig,
                          cluster_utils.get_hacluster_config)

    @patch.object(cluster_utils, 'valid_hacluster_config')
    def test_get_hacluster_config_with_excludes(self, valid_hacluster_config):
        '''It fetches all hacluster charm config'''
        conf = {
            'ha-bindiface': 'eth1',
            'ha-mcastport': '3333',
        }
        valid_hacluster_config.return_value = True

        def _fake_config_get(setting):
            return conf[setting]

        self.config_get.side_effect = _fake_config_get
        exclude_keys = ['vip', 'os-admin-hostname', 'os-internal-hostname',
                        'os-public-hostname', 'os-access-hostname']
        result = cluster_utils.get_hacluster_config(exclude_keys)
        self.assertEquals(conf, result)

    @patch.object(cluster_utils, 'is_clustered')
    def test_canonical_url_bare(self, is_clustered):
        '''It constructs a URL to host with no https or cluster present'''
        self.unit_get.return_value = 'foohost1'
        is_clustered.return_value = False
        configs = MagicMock()
        configs.complete_contexts = MagicMock()
        configs.complete_contexts.return_value = []
        url = cluster_utils.canonical_url(configs)
        self.assertEquals('http://foohost1', url)

    @patch.object(cluster_utils, 'is_clustered')
    def test_canonical_url_https_no_cluster(self, is_clustered):
        '''It constructs a URL to host with https and no cluster present'''
        self.unit_get.return_value = 'foohost1'
        is_clustered.return_value = False
        configs = MagicMock()
        configs.complete_contexts = MagicMock()
        configs.complete_contexts.return_value = ['https']
        url = cluster_utils.canonical_url(configs)
        self.assertEquals('https://foohost1', url)

    @patch.object(cluster_utils, 'is_clustered')
    def test_canonical_url_https_cluster(self, is_clustered):
        '''It constructs a URL to host with https and cluster present'''
        self.config_get.return_value = '10.0.0.1'
        is_clustered.return_value = True
        configs = MagicMock()
        configs.complete_contexts = MagicMock()
        configs.complete_contexts.return_value = ['https']
        url = cluster_utils.canonical_url(configs)
        self.assertEquals('https://10.0.0.1', url)

    @patch.object(cluster_utils, 'is_clustered')
    def test_canonical_url_cluster_no_https(self, is_clustered):
        '''It constructs a URL to host with no https and cluster present'''
        self.config_get.return_value = '10.0.0.1'
        self.unit_get.return_value = 'foohost1'
        is_clustered.return_value = True
        configs = MagicMock()
        configs.complete_contexts = MagicMock()
        configs.complete_contexts.return_value = []
        url = cluster_utils.canonical_url(configs)
        self.assertEquals('http://10.0.0.1', url)

    @patch.object(cluster_utils, 'status_set')
    def test_valid_hacluster_config_incomplete(self, status_set):
        '''Returns False with incomplete HA config'''
        conf = {
            'vip': None,
            'os-admin-hostname': None,
            'os-public-hostname': None,
            'os-internal-hostname': None,
            'os-access-hostname': None,
            'dns-ha': False,
        }

        def _fake_config_get(setting):
            return conf[setting]

        self.config_get.side_effect = _fake_config_get
        self.assertRaises(cluster_utils.HAIncorrectConfig,
                          cluster_utils.valid_hacluster_config)

    @patch.object(cluster_utils, 'status_set')
    def test_valid_hacluster_config_both(self, status_set):
        '''Returns False when both VIP and DNS HA are set'''
        conf = {
            'vip': '10.0.0.1',
            'os-admin-hostname': None,
            'os-public-hostname': None,
            'os-internal-hostname': None,
            'os-access-hostname': None,
            'dns-ha': True,
        }

        def _fake_config_get(setting):
            return conf[setting]

        self.config_get.side_effect = _fake_config_get
        self.assertRaises(cluster_utils.HAIncorrectConfig,
                          cluster_utils.valid_hacluster_config)

    @patch.object(cluster_utils, 'status_set')
    def test_valid_hacluster_config_vip_ha(self, status_set):
        '''Returns True with complete VIP HA config'''
        conf = {
            'vip': '10.0.0.1',
            'os-admin-hostname': None,
            'os-public-hostname': None,
            'os-internal-hostname': None,
            'os-access-hostname': None,
            'dns-ha': False,
        }

        def _fake_config_get(setting):
            return conf[setting]

        self.config_get.side_effect = _fake_config_get
        self.assertTrue(cluster_utils.valid_hacluster_config())
        self.assertFalse(status_set.called)

    @patch.object(cluster_utils, 'status_set')
    def test_valid_hacluster_config_dns_incomplete(self, status_set):
        '''Returns False with incomplete DNS HA config'''
        conf = {
            'vip': None,
            'os-admin-hostname': None,
            'os-public-hostname': None,
            'os-internal-hostname': None,
            'os-access-hostname': None,
            'dns-ha': True,
        }

        def _fake_config_get(setting):
            return conf[setting]

        self.config_get.side_effect = _fake_config_get
        self.assertRaises(cluster_utils.HAIncompleteConfig,
                          cluster_utils.valid_hacluster_config)

    @patch.object(cluster_utils, 'status_set')
    def test_valid_hacluster_config_dns_ha(self, status_set):
        '''Returns True with complete DNS HA config'''
        conf = {
            'vip': None,
            'os-admin-hostname': 'somehostname',
            'os-public-hostname': None,
            'os-internal-hostname': None,
            'os-access-hostname': None,
            'dns-ha': True,
        }

        def _fake_config_get(setting):
            return conf[setting]

        self.config_get.side_effect = _fake_config_get
        self.assertTrue(cluster_utils.valid_hacluster_config())
        self.assertFalse(status_set.called)

    @patch.object(cluster_utils, 'juju_is_leader')
    @patch.object(cluster_utils, 'status_set')
    @patch.object(cluster_utils.time, 'sleep')
    @patch.object(cluster_utils, 'modulo_distribution')
    @patch.object(cluster_utils, 'log')
    def test_distributed_wait(self, log, modulo_distribution, sleep,
                              status_set, is_leader):

        # Leader regardless of modulo should not wait
        is_leader.return_value = True
        cluster_utils.distributed_wait(modulo=9, wait=23)
        modulo_distribution.assert_not_called()
        sleep.assert_called_with(0)

        # The rest of the tests are non-leader units
        is_leader.return_value = False

        def _fake_config_get(setting):
            return conf[setting]

        # Uses fallback defaults
        conf = {
            'modulo-nodes': None,
            'known-wait': None,
        }
        self.config_get.side_effect = _fake_config_get
        cluster_utils.distributed_wait()
        modulo_distribution.assert_called_with(modulo=3, wait=30,
                                               non_zero_wait=True)

        # Uses config values
        conf = {
            'modulo-nodes': 7,
            'known-wait': 10,
        }
        self.config_get.side_effect = _fake_config_get
        cluster_utils.distributed_wait()
        modulo_distribution.assert_called_with(modulo=7, wait=10,
                                               non_zero_wait=True)

        # Uses passed values
        cluster_utils.distributed_wait(modulo=5, wait=45)
        modulo_distribution.assert_called_with(modulo=5, wait=45,
                                               non_zero_wait=True)

    @patch.object(cluster_utils, 'relation_ids')
    def test_get_managed_services_and_ports(self, relation_ids):
        relation_ids.return_value = ['rel:2']
        self.assertEqual(
            cluster_utils.get_managed_services_and_ports(
                ['apache2', 'haproxy'],
                [8067, 4545, 6732]),
            (['apache2'], [8057, 4535, 6722]))
        self.assertEqual(
            cluster_utils.get_managed_services_and_ports(
                ['apache2', 'haproxy'],
                [8067, 4545, 6732],
                external_services=['apache2']),
            (['haproxy'], [8057, 4535, 6722]))

        def add_ten(x):
            return x + 10

        self.assertEqual(
            cluster_utils.get_managed_services_and_ports(
                ['apache2', 'haproxy'],
                [8067, 4545, 6732],
                port_conv_f=add_ten),
            (['apache2'], [8077, 4555, 6742]))



# Copyright 2016 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from charmhelpers.core.hookenv import (
    log,
    DEBUG,
    WARNING,
)

try:
    from jinja2 import FileSystemLoader, Environment
except ImportError:
    from charmhelpers.fetch import apt_install
    from charmhelpers.fetch import apt_update
    apt_update(fatal=True)
    apt_install('python3-jinja2', fatal=True)
    from jinja2 import FileSystemLoader, Environment


# NOTE: function separated from main rendering code to facilitate easier
#       mocking in unit tests.
def write(path, data):
    with open(path, 'wb') as out:
        out.write(data)


def get_template_path(template_dir, path):
    """Returns the template file which would be used to render the path.

    The path to the template file is returned.
    :param template_dir: the directory the templates are located in
    :param path: the file path to be written to.
    :returns: path to the template file
    """
    return os.path.join(template_dir, os.path.basename(path))


def render_and_write(template_dir, path, context):
    """Renders the specified template into the file.

    :param template_dir: the directory to load the template from
    :param path: the path to write the templated contents to
    :param context: the parameters to pass to the rendering engine
    """
    env = Environment(loader=FileSystemLoader(template_dir))
    template_file = os.path.basename(path)
    template = env.get_template(template_file)
    log('Rendering from template: %s' % template.name, level=DEBUG)
    rendered_content = template.render(context)
    if not rendered_content:
        log("Render returned None - skipping '%s'" % path,
            level=WARNING)
        return

    write(path, rendered_content.encode('utf-8').strip())
    log('Wrote template %s' % path, level=DEBUG)



#!/usr/bin/env python3
#
# Copyright 2016-2022 Canonical Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import glob
import itertools
import logging
import os
import re
import shlex
import subprocess
import typing


def format_pci_addr(pci_addr: str) -> str:
    """Format a PCI address with 0 fill for parts

    :param: pci_addr: unformatted PCI address
    :type: str
    :returns: formatted PCI address
    :rtype: str
    """
    domain, bus, slot_func = pci_addr.split(":")
    slot, func = slot_func.split(".")
    return "{}:{}:{}.{}".format(
        domain.zfill(4), bus.zfill(2), slot.zfill(2), func
    )


def get_sysnet_interfaces_and_macs() -> list:
    """Catalog interface information from local system

    each device dict contains:

        interface: logical name
        mac_address: MAC address
        pci_address: PCI address
        state: Current interface state (up/down)
        sriov: Boolean indicating whether interface is an SR-IOV
               capable device.
        sriov_totalvfs: Total VF capacity of device
        sriov_numvfs: Configured VF capacity of device

    :returns: array of dict objects containing details of each interface
    :rtype: list
    """
    net_devs = []
    for sdir in itertools.chain(
            glob.glob("/sys/bus/pci/devices/*/net/../"),
            glob.glob("/sys/bus/pci/devices/*/virtio*/net/../")):
        fq_path = os.path.realpath(sdir)
        path = fq_path.split("/")
        if "virtio" in path[-1]:
            pci_address = path[-2]
        else:
            pci_address = path[-1]
        ifname = get_sysnet_interface(sdir)
        if not ifname:
            logging.warn("Unable to determine interface name for PCI "
                         "device {}".format(pci_address))
            continue
        device = {
            "interface": ifname,
            "mac_address": get_sysnet_mac(sdir, ifname),
            "pci_address": pci_address,
            "state": get_sysnet_device_state(sdir, ifname),
            "sriov": is_sriov(sdir),
        }
        if device["sriov"]:
            device["sriov_totalvfs"] = get_sriov_totalvfs(sdir)
            device["sriov_numvfs"] = get_sriov_numvfs(sdir)
        net_devs.append(device)

    return net_devs


def get_sysnet_mac(sysdir: str, ifname: str) -> str:
    """Determine MAC address for a device

    :param: sysdir: path to device /sys directory
    :type: str
    :returns: MAC address of device
    :rtype: str
    """
    mac_addr_file = os.path.join(sysdir, "net", ifname, "address")
    with open(mac_addr_file, "r") as f:
        read_data = f.read()
    return read_data.strip()


def get_sysnet_device_state(sysdir: str, ifname: str) -> str:
    """Read operational state of a device

    :param: sysdir: path to device /sys directory
    :type: str
    :returns: current device state
    :rtype: str
    """
    state_file = os.path.join(sysdir, "net", ifname, "operstate")
    with open(state_file, "r") as f:
        read_data = f.read()
    return read_data.strip()


def is_sriov(sysdir: str) -> bool:
    """Determine whether a device is SR-IOV capable

    :param: sysdir: path to device /sys directory
    :type: str
    :returns: whether device is SR-IOV capable or not
    :rtype: bool
    """
    return os.path.exists(os.path.join(sysdir, "sriov_totalvfs"))


def get_sriov_totalvfs(sysdir: str) -> int:
    """Read total VF capacity for a device

    :param: sysdir: path to device /sys directory
    :type: str
    :returns: number of VF's the device supports
    :rtype: int
    """
    sriov_totalvfs_file = os.path.join(sysdir, "sriov_totalvfs")
    with open(sriov_totalvfs_file, "r") as f:
        read_data = f.read()
    return int(read_data.strip())


def get_sriov_numvfs(sysdir: str) -> int:
    """Read configured VF capacity for a device

    :param: sysdir: path to device /sys directory
    :type: str
    :returns: number of VF's the device is configured with
    :rtype: int
    """
    sriov_numvfs_file = os.path.join(sysdir, "sriov_numvfs")
    with open(sriov_numvfs_file, "r") as f:
        read_data = f.read()
    return int(read_data.strip())


# https://github.com/libvirt/libvirt/commit/5b1c525b1f3608156884aed0dc5e925306c1e260
PF_PHYS_PORT_NAME_REGEX = re.compile(r"(p[0-9]+$)|(p[0-9]+s[0-9]+$)",
                                     re.IGNORECASE)


def _phys_port_name_is_pf(sysnetdir: str) -> typing.Optional[bool]:
    try:
        with open(os.path.join(sysnetdir, "phys_port_name"), "r") as fin:
            return (PF_PHYS_PORT_NAME_REGEX.match(fin.read().strip())
                    is not None)
    except OSError:
        return


def get_sysnet_interface(sysdir: str) -> typing.Optional[str]:
    sysnetdir = os.path.join(sysdir, "net")
    netdevs = os.listdir(sysnetdir)
    # Return early in case the PCI device only has one netdev
    if len(netdevs) == 1:
        return netdevs[0]

    # When a PCI device has multiple netdevs we need to figure out which one
    # represents the PF
    for netdev in netdevs:
        if _phys_port_name_is_pf(os.path.join(sysnetdir, netdev)):
            return netdev


def get_pci_ethernet_addresses() -> list:
    """Generate list of PCI addresses for all network adapters

    :returns: list of PCI addresses
    :rtype: list
    """
    cmd = ["lspci", "-m", "-D"]
    lspci_output = subprocess.check_output(cmd).decode("UTF-8")
    pci_addresses = []
    for line in lspci_output.split("\n"):
        columns = shlex.split(line)
        if len(columns) > 1 and columns[1] == "Ethernet controller":
            pci_address = columns[0]
            pci_addresses.append(format_pci_addr(pci_address))
    return pci_addresses


class PCINetDevice(object):
    def __init__(self, pci_address):
        self.pci_address = pci_address
        self.interface_name = None
        self.mac_address = None
        self.state = None
        self.sriov = False
        self.sriov_totalvfs = None
        self.sriov_numvfs = None
        self.update_attributes()

    def update_attributes(self):
        self.update_interface_info()

    def update_interface_info(self):
        net_devices = get_sysnet_interfaces_and_macs()
        for interface in net_devices:
            if self.pci_address == interface["pci_address"]:
                self.interface_name = interface["interface"]
                self.mac_address = interface["mac_address"]
                self.state = interface["state"]
                self.sriov = interface["sriov"]
                if self.sriov:
                    self.sriov_totalvfs = interface["sriov_totalvfs"]
                    self.sriov_numvfs = interface["sriov_numvfs"]

    def _set_sriov_numvfs(self, numvfs: int):
        sdevice = os.path.join(
            "/sys/bus/pci/devices", self.pci_address, "sriov_numvfs"
        )
        with open(sdevice, "w") as sh:
            sh.write(str(numvfs))
        self.update_attributes()

    def set_sriov_numvfs(self, numvfs: int) -> bool:
        """Set the number of VF devices for a SR-IOV PF

        Assuming the device is an SR-IOV device, this function will attempt
        to change the number of VF's created by the PF.

        @param numvfs: integer to set the current number of VF's to
        @returns boolean indicating whether any changes where made
        """
        if self.sriov and numvfs != self.sriov_numvfs:
            # NOTE(fnordahl): run-time change of numvfs is disallowed
            # without resetting to 0 first.
            self._set_sriov_numvfs(0)
            self._set_sriov_numvfs(numvfs)
            return True
        return False


class PCINetDevices(object):
    def __init__(self):
        self.pci_devices = [
            PCINetDevice(dev) for dev in get_pci_ethernet_addresses()
        ]

    def update_devices(self):
        for pcidev in self.pci_devices:
            pcidev.update_attributes()

    def get_macs(self) -> list:
        macs = []
        for pcidev in self.pci_devices:
            if pcidev.mac_address:
                macs.append(pcidev.mac_address)
        return macs

    def get_device_from_mac(self, mac: str) -> PCINetDevice:
        for pcidev in self.pci_devices:
            if pcidev.mac_address == mac:
                return pcidev
        return None

    def get_device_from_pci_address(self, pci_addr: str) -> PCINetDevice:
        for pcidev in self.pci_devices:
            if pcidev.pci_address == pci_addr:
                return pcidev
        return None

    def get_device_from_interface_name(
        self, interface_name: str
    ) -> PCINetDevice:
        for pcidev in self.pci_devices:
            if pcidev.interface_name == interface_name:
                return pcidev
        return None



import mock
import subprocess
import unittest

from tests.helpers import patch_open, mock_open

import charmhelpers.contrib.openstack.ssh_migrations as ssh_migrations


UNIT1_HOST_KEY_1 = """|1|EaIiWNsBsaSke5T5bdDlaV5xKPU=|WKMu3Va+oNwRjXmPGOZ+mrpWbM8= ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQDZdZdR7I35ymFdspruN1CIez/0m62sJeld2nLuOGaNbdl/rk5bGrWUAZh6c9p9H53FAqGAXBD/1C8dZ5dgIAGdTs7PAZq7owXCpgUPQcGOYVAtBwv8qfnWyI1W+Vpi6vnb2sgYr6XGbB9b84i4vrd98IIpXIleC9qd0VUTSYgd7+NPaFNoK0HZmqcNEf5leaa8sgSf4t5F+BTWEXzU3ql/3isFT8lEpJ9N8wOvNzAoFEQcxqauvOJn72QQ6kUrQT3NdQFUMHquS/s+nBrQNPbUmzqrvSOed75Qk8359zqU1Rce7U39cqc0scYi1ak3oJdojwfLFKJw4TMPn/Pq7JnT"""
UNIT1_HOST_KEY_2 = """|1|mCyYWqJl8loqV6LCY84lu2rpqLA=|51m+M+0ES3jYVzr3Kco3CDg8hEY= ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQDZdZdR7I35ymFdspruN1CIez/0m62sJeld2nLuOGaNbdl/rk5bGrWUAZh6c9p9H53FAqGAXBD/1C8dZ5dgIAGdTs7PAZq7owXCpgUPQcGOYVAtBwv8qfnWyI1W+Vpi6vnb2sgYr6XGbB9b84i4vrd98IIpXIleC9qd0VUTSYgd7+NPaFNoK0HZmqcNEf5leaa8sgSf4t5F+BTWEXzU3ql/3isFT8lEpJ9N8wOvNzAoFEQcxqauvOJn72QQ6kUrQT3NdQFUMHquS/s+nBrQNPbUmzqrvSOed75Qk8359zqU1Rce7U39cqc0scYi1ak3oJdojwfLFKJw4TMPn/Pq7JnT"""
UNIT2_HOST_KEY_1 = """|1|eWagMqrN7XmX7NdVpZbqMZ2cb4Q=|3jgGiFEU9SMhXwdX0w0kkG54CZc= ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQC27Lv4wtAiPIOTsrUCFOU4qaNsov+LZSVHtxlu0aERuD+oU3ZILXOITXJlDohweXN6YuP4hFg49FF119gmMag+hDiA8/BztQmsplkwHrWEPuWKpReLRNLBU+Nt78jrJkjTK8Egwxbaxu8fAPZkCgGyeLkIH4ghrdWlOaWYXzwuxXkYWSpQOgF6E/T+19JKVKNpt2i6w7q9vVwZEjwVr30ubs1bNdPzE9ylNLQRrGa7c38SKsEos5RtZJjEuZGTC9KI0QdEgwnxxNMlT/CIgwWA1V38vLsosF2pHKxbmtCNvBuPNtrBDgXhukVqyEh825RhTAyQYGshMCbAbxl/M9c3"""
UNIT2_HOST_KEY_2 = """|1|zRH8troNwhVzrkMx86E5Ibevw5s=|gESlgkwUumP8q0A6l+CoRlFRpTw= ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQC27Lv4wtAiPIOTsrUCFOU4qaNsov+LZSVHtxlu0aERuD+oU3ZILXOITXJlDohweXN6YuP4hFg49FF119gmMag+hDiA8/BztQmsplkwHrWEPuWKpReLRNLBU+Nt78jrJkjTK8Egwxbaxu8fAPZkCgGyeLkIH4ghrdWlOaWYXzwuxXkYWSpQOgF6E/T+19JKVKNpt2i6w7q9vVwZEjwVr30ubs1bNdPzE9ylNLQRrGa7c38SKsEos5RtZJjEuZGTC9KI0QdEgwnxxNMlT/CIgwWA1V38vLsosF2pHKxbmtCNvBuPNtrBDgXhukVqyEh825RhTAyQYGshMCbAbxl/M9c3"""
HOST_KEYS = [UNIT1_HOST_KEY_1, UNIT1_HOST_KEY_2,
             UNIT2_HOST_KEY_1, UNIT2_HOST_KEY_2]
UNIT1_PUBKEY_1 = """ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQDqtnB3zh3sMufZd1khi44su0hTg/LqLb3ma2iyueTULZikDYa65UidVxzsa6r0Y9jkHwknGlh7fNnGdmc3S8EE+rVUNF4r3JF2Zd/pdfCBia/BmKJcO7+NyRWc8ihlrA3xYUSm+Yg8ZIpqoSb1LKjgAdYISh9HQQaXut2sXtHESdpilNpDf42AZfuQM+B0op0v7bq86ZXOM1rvdJriI6BduHaAOux+d9HDNvV5AxYTICrUkXqIvdHnoRyOFfhTcKun0EtuUxpDiAi0im9C+i+MPwMvA6AmRbot6Tqt2xZPRBYY8+WF7I5cBoovES/dWKP5TZwaGBr+WNv+z2JJhvlN root@juju-4665be-20180716142533-8"""
UNIT2_PUBKEY_1 = """ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQCkWfkVrG7wTnfifvL0GkmDj6L33PKrWjzN2hOwZb9EoxwNzFGTMTBIpepTAnO6hdFBwtus1Ej/L12K6L/0YRDZAKjE7yTWOsh1kUxPZ1INRCqLILiefE5A/LPNx8NDb+d/2ryc5QmOQXUALs6mC5VDNchImUp9L7l0RIzPOgPXZCqMC1nZLqqX+eI9EUaf29/+NztYw59rFAa3hWNe8RJCSFeU+iWirWP8rfX9jsLzD9hO3nuZjP23M6tv1jX9LQD+8qkx0WSMa2WrIjkMiclP6tkyCJOZogyoPzZm/+dUhLeY9bIizbZCQKH/b4gOl5m/PkWoqEFshfqGzUIPkAJp root@juju-4665be-20180716142533-9"""
PUB_KEYS = [UNIT1_PUBKEY_1, UNIT2_PUBKEY_1]


class SSHMigrationsTests(unittest.TestCase):

    def setUp(self):
        self._patches = {}
        self._patches_start = {}

    def tearDown(self):
        """Run teardown of patches."""
        for k, v in self._patches.items():
            v.stop()
            setattr(self, k, None)
        self._patches = None
        self._patches_start = None

    def patch_object(self, obj, attr, return_value=None, name=None, new=None,
                     **kwargs):
        """Patch the given object."""
        if name is None:
            name = attr
        if new is not None:
            mocked = mock.patch.object(obj, attr, new=new, **kwargs)
        else:
            mocked = mock.patch.object(obj, attr, **kwargs)
        self._patches[name] = mocked
        started = mocked.start()
        if new is None:
            started.return_value = return_value
        self._patches_start[name] = started
        setattr(self, name, started)

    def setup_mocks_ssh_directory_for_unit(self, app_name, ssh_dir_exists,
                                           app_dir_exists, auth_keys_exists,
                                           known_hosts_exists, user=None):
        def _isdir(x):
            return {
                ssh_dir + '/': ssh_dir_exists,
                app_dir: app_dir_exists}[x]

        def _isfile(x):
            return {
                '{}/authorized_keys'.format(app_dir): auth_keys_exists,
                '{}/known_hosts'.format(app_dir): known_hosts_exists}[x]

        if user:
            app_name = "{}_{}".format(app_name, user)
        ssh_dir = '/etc/nova/compute_ssh'
        app_dir = '{}/{}'.format(ssh_dir, app_name)

        self.patch_object(ssh_migrations.os, 'mkdir')
        self.patch_object(ssh_migrations.os.path, 'isdir', side_effect=_isdir)
        self.patch_object(ssh_migrations.os.path, 'isfile',
                          side_effect=_isfile)

    def test_ssh_directory_for_unit(self):
        self.setup_mocks_ssh_directory_for_unit(
            'nova-compute-lxd',
            ssh_dir_exists=True,
            app_dir_exists=True,
            auth_keys_exists=True,
            known_hosts_exists=True)
        self.assertEqual(
            ssh_migrations.ssh_directory_for_unit('nova-compute-lxd'),
            '/etc/nova/compute_ssh/nova-compute-lxd')
        self.assertFalse(self.mkdir.called)

    def test_ssh_directory_for_unit_user(self):
        self.setup_mocks_ssh_directory_for_unit(
            'nova-compute-lxd',
            ssh_dir_exists=True,
            app_dir_exists=True,
            auth_keys_exists=True,
            known_hosts_exists=True,
            user='nova')
        self.assertEqual(
            ssh_migrations.ssh_directory_for_unit(
                'nova-compute-lxd',
                user='nova'),
            '/etc/nova/compute_ssh/nova-compute-lxd_nova')
        self.assertFalse(self.mkdir.called)

    def test_ssh_directory_missing_dir(self):
        self.setup_mocks_ssh_directory_for_unit(
            'nova-compute-lxd',
            ssh_dir_exists=False,
            app_dir_exists=True,
            auth_keys_exists=True,
            known_hosts_exists=True)
        self.assertEqual(
            ssh_migrations.ssh_directory_for_unit('nova-compute-lxd'),
            '/etc/nova/compute_ssh/nova-compute-lxd')
        self.mkdir.assert_called_once_with('/etc/nova/compute_ssh/')

    def test_ssh_directory_missing_dirs(self):
        self.setup_mocks_ssh_directory_for_unit(
            'nova-compute-lxd',
            ssh_dir_exists=False,
            app_dir_exists=False,
            auth_keys_exists=True,
            known_hosts_exists=True)
        self.assertEqual(
            ssh_migrations.ssh_directory_for_unit('nova-compute-lxd'),
            '/etc/nova/compute_ssh/nova-compute-lxd')
        mkdir_calls = [
            mock.call('/etc/nova/compute_ssh/'),
            mock.call('/etc/nova/compute_ssh/nova-compute-lxd')]
        self.mkdir.assert_has_calls(mkdir_calls)

    @mock.patch('builtins.open')
    def test_ssh_directory_missing_file(self, _open):
        self.setup_mocks_ssh_directory_for_unit(
            'nova-compute-lxd',
            ssh_dir_exists=True,
            app_dir_exists=True,
            auth_keys_exists=False,
            known_hosts_exists=True)
        self.assertEqual(
            ssh_migrations.ssh_directory_for_unit('nova-compute-lxd'),
            '/etc/nova/compute_ssh/nova-compute-lxd')
        _open.assert_called_once_with(
            '/etc/nova/compute_ssh/nova-compute-lxd/authorized_keys',
            'w')
        self.assertFalse(self.mkdir.called)

    @mock.patch('builtins.open')
    def test_ssh_directory_missing_files(self, _open):
        self.setup_mocks_ssh_directory_for_unit(
            'nova-compute-lxd',
            ssh_dir_exists=True,
            app_dir_exists=True,
            auth_keys_exists=False,
            known_hosts_exists=False)
        self.assertEqual(
            ssh_migrations.ssh_directory_for_unit('nova-compute-lxd'),
            '/etc/nova/compute_ssh/nova-compute-lxd')
        open_calls = [
            mock.call(
                '/etc/nova/compute_ssh/nova-compute-lxd/authorized_keys',
                'w'),
            mock.call().close(),
            mock.call(
                '/etc/nova/compute_ssh/nova-compute-lxd/known_hosts',
                'w'),
            mock.call().close()]
        _open.assert_has_calls(open_calls)
        self.assertFalse(self.mkdir.called)

    def setup_ssh_directory_for_unit_mocks(self):
        self.patch_object(
            ssh_migrations,
            'ssh_directory_for_unit',
            return_value='/somedir')

    def test_known_hosts(self):
        self.setup_ssh_directory_for_unit_mocks()
        self.assertEqual(
            ssh_migrations.known_hosts('nova-compute-lxd'),
            '/somedir/known_hosts')

    def test_authorized_keys(self):
        self.setup_ssh_directory_for_unit_mocks()
        self.assertEqual(
            ssh_migrations.authorized_keys('nova-compute-lxd'),
            '/somedir/authorized_keys')

    @mock.patch('subprocess.check_output')
    def test_ssh_known_host_key(self, _check_output):
        self.setup_ssh_directory_for_unit_mocks()
        _check_output.return_value = UNIT1_HOST_KEY_1
        self.assertEqual(
            ssh_migrations.ssh_known_host_key(
                'juju-4665be-20180716142533-8',
                'nova-compute-lxd'),
            UNIT1_HOST_KEY_1)

    @mock.patch('subprocess.check_output')
    def test_ssh_known_host_key_multi_match(self, _check_output):
        self.setup_ssh_directory_for_unit_mocks()
        _check_output.return_value = '{}\n{}\n'.format(UNIT1_HOST_KEY_1,
                                                       UNIT1_HOST_KEY_2)
        self.assertEqual(
            ssh_migrations.ssh_known_host_key(
                'juju-4665be-20180716142533-8',
                'nova-compute-lxd'),
            UNIT1_HOST_KEY_1)

    @mock.patch('subprocess.check_output')
    def test_ssh_known_host_key_rc1(self, _check_output):
        self.setup_ssh_directory_for_unit_mocks()
        _check_output.side_effect = subprocess.CalledProcessError(
            cmd=['anything'],
            returncode=1,
            output=UNIT1_HOST_KEY_1)
        self.assertEqual(
            ssh_migrations.ssh_known_host_key(
                'juju-4665be-20180716142533-8',
                'nova-compute-lxd'),
            UNIT1_HOST_KEY_1)

    @mock.patch('subprocess.check_output')
    def test_ssh_known_host_key_rc2(self, _check_output):
        self.setup_ssh_directory_for_unit_mocks()
        _check_output.side_effect = subprocess.CalledProcessError(
            cmd=['anything'],
            returncode=2,
            output='')
        with self.assertRaises(subprocess.CalledProcessError):
            ssh_migrations.ssh_known_host_key(
                'juju-4665be-20180716142533-8',
                'nova-compute-lxd')

    @mock.patch('subprocess.check_output')
    def test_ssh_known_host_key_no_match(self, _check_output):
        self.setup_ssh_directory_for_unit_mocks()
        _check_output.return_value = ''
        self.assertIsNone(
            ssh_migrations.ssh_known_host_key(
                'juju-4665be-20180716142533-8',
                'nova-compute-lxd'))

    @mock.patch('subprocess.check_call')
    def test_remove_known_host(self, _check_call):
        self.patch_object(ssh_migrations, 'log')
        self.setup_ssh_directory_for_unit_mocks()
        ssh_migrations.remove_known_host(
            'juju-4665be-20180716142533-8',
            'nova-compute-lxd')
        _check_call.assert_called_once_with([
            'ssh-keygen',
            '-f',
            '/somedir/known_hosts',
            '-R',
            'juju-4665be-20180716142533-8'])

    def test_is_same_key(self):
        self.assertTrue(
            ssh_migrations.is_same_key(UNIT1_HOST_KEY_1, UNIT1_HOST_KEY_2))

    def test_is_same_key_false(self):
        self.assertFalse(
            ssh_migrations.is_same_key(UNIT1_HOST_KEY_1, UNIT2_HOST_KEY_1))

    def setup_mocks_add_known_host(self):
        self.setup_ssh_directory_for_unit_mocks()
        self.patch_object(ssh_migrations.subprocess, 'check_output')
        self.patch_object(ssh_migrations, 'log')
        self.patch_object(ssh_migrations, 'ssh_known_host_key')
        self.patch_object(ssh_migrations, 'remove_known_host')

    def test_add_known_host(self):
        self.setup_mocks_add_known_host()
        self.check_output.return_value = UNIT1_HOST_KEY_1
        self.ssh_known_host_key.return_value = ''
        with patch_open() as (mock_open, mock_file):
            ssh_migrations.add_known_host(
                'juju-4665be-20180716142533-8',
                'nova-compute-lxd')
            mock_file.write.assert_called_with(UNIT1_HOST_KEY_1 + '\n')
            mock_open.assert_called_with('/somedir/known_hosts', 'a')
        self.assertFalse(self.remove_known_host.called)

    def test_add_known_host_existing_invalid_key(self):
        self.setup_mocks_add_known_host()
        self.check_output.return_value = UNIT1_HOST_KEY_1
        self.ssh_known_host_key.return_value = UNIT2_HOST_KEY_1
        with patch_open() as (mock_open, mock_file):
            ssh_migrations.add_known_host(
                'juju-4665be-20180716142533-8',
                'nova-compute-lxd')
            mock_file.write.assert_called_with(UNIT1_HOST_KEY_1 + '\n')
            mock_open.assert_called_with('/somedir/known_hosts', 'a')
        self.remove_known_host.assert_called_once_wth(
            'juju-4665be-20180716142533-8',
            'nova-compute-lxd')

    def test_add_known_host_existing_valid_key(self):
        self.setup_mocks_add_known_host()
        self.check_output.return_value = UNIT2_HOST_KEY_1
        self.ssh_known_host_key.return_value = UNIT2_HOST_KEY_1
        with patch_open() as (mock_open, mock_file):
            ssh_migrations.add_known_host(
                'juju-4665be-20180716142533-8',
                'nova-compute-lxd')
            self.assertFalse(mock_open.called)
        self.assertFalse(self.remove_known_host.called)

    def test_ssh_authorized_key_exists(self):
        self.setup_mocks_add_known_host()
        contents = '{}\n{}\n'.format(UNIT1_PUBKEY_1, UNIT2_PUBKEY_1)
        with mock_open('/somedir/authorized_keys', contents=contents):
            self.assertTrue(
                ssh_migrations.ssh_authorized_key_exists(
                    UNIT1_PUBKEY_1,
                    'nova-compute-lxd'))

    def test_ssh_authorized_key_exists_false(self):
        self.setup_mocks_add_known_host()
        contents = '{}\n'.format(UNIT1_PUBKEY_1)
        with mock_open('/somedir/authorized_keys', contents=contents):
            self.assertFalse(
                ssh_migrations.ssh_authorized_key_exists(
                    UNIT2_PUBKEY_1,
                    'nova-compute-lxd'))

    def test_add_authorized_key(self):
        self.setup_mocks_add_known_host()
        with patch_open() as (mock_open, mock_file):
            ssh_migrations.add_authorized_key(
                UNIT1_PUBKEY_1,
                'nova-compute-lxd')
            mock_file.write.assert_called_with(UNIT1_PUBKEY_1 + '\n')
            mock_open.assert_called_with('/somedir/authorized_keys', 'a')

    def setup_mocks_ssh_compute_add_host_and_key(self):
        self.setup_ssh_directory_for_unit_mocks()
        self.patch_object(ssh_migrations, 'log')
        self.patch_object(ssh_migrations, 'get_hostname')
        self.patch_object(ssh_migrations, 'get_host_ip')
        self.patch_object(ssh_migrations, 'ns_query')
        self.patch_object(ssh_migrations, 'add_known_host')
        self.patch_object(ssh_migrations, 'ssh_authorized_key_exists')
        self.patch_object(ssh_migrations, 'add_authorized_key')

    def test_ssh_compute_add_host_and_key(self):
        self.setup_mocks_ssh_compute_add_host_and_key()
        self.get_hostname.return_value = 'alt-hostname.project.serverstack'
        self.ns_query.return_value = '10.6.0.17'
        ssh_migrations.ssh_compute_add_host_and_key(
            UNIT1_PUBKEY_1,
            'juju-4665be-20180716142533-8.project.serverstack',
            '10.5.0.17',
            'nova-compute-lxd')
        expect_hosts = [
            'juju-4665be-20180716142533-8.project.serverstack',
            'alt-hostname.project.serverstack',
            'alt-hostname']
        add_known_host_calls = []
        for host in expect_hosts:
            add_known_host_calls.append(
                mock.call(host, 'nova-compute-lxd', None))
        self.add_known_host.assert_has_calls(
            add_known_host_calls,
            any_order=True)
        self.add_authorized_key.assert_called_once_with(
            UNIT1_PUBKEY_1,
            'nova-compute-lxd',
            None)

    def test_ssh_compute_add_host_and_key_priv_addr_not_ip(self):
        self.setup_mocks_ssh_compute_add_host_and_key()
        self.get_hostname.return_value = 'alt-hostname.project.serverstack'
        self.ns_query.return_value = '10.6.0.17'
        self.get_host_ip.return_value = '10.6.0.17'
        ssh_migrations.ssh_compute_add_host_and_key(
            UNIT1_PUBKEY_1,
            'juju-4665be-20180716142533-8.project.serverstack',
            'bob.maas',
            'nova-compute-lxd')
        expect_hosts = [
            'bob.maas',
            'juju-4665be-20180716142533-8.project.serverstack',
            '10.6.0.17',
            'bob']
        add_known_host_calls = []
        for host in expect_hosts:
            add_known_host_calls.append(
                mock.call(host, 'nova-compute-lxd', None))
        self.add_known_host.assert_has_calls(
            add_known_host_calls,
            any_order=True)
        self.add_authorized_key.assert_called_once_with(
            UNIT1_PUBKEY_1,
            'nova-compute-lxd',
            None)

    def test_ssh_compute_add_host_and_key_ipv6(self):
        self.setup_mocks_ssh_compute_add_host_and_key()
        ssh_migrations.ssh_compute_add_host_and_key(
            UNIT1_PUBKEY_1,
            'juju-4665be-20180716142533-8.project.serverstack',
            'fe80::8842:a9ff:fe53:72e4',
            'nova-compute-lxd')
        self.add_known_host.assert_called_once_with(
            'fe80::8842:a9ff:fe53:72e4',
            'nova-compute-lxd',
            None)
        self.add_authorized_key.assert_called_once_with(
            UNIT1_PUBKEY_1,
            'nova-compute-lxd',
            None)

    @mock.patch.object(ssh_migrations, 'ssh_compute_add_host_and_key')
    @mock.patch.object(ssh_migrations, 'relation_get')
    def test_ssh_compute_add(self, _relation_get,
                             _ssh_compute_add_host_and_key):
        _relation_get.return_value = {
            'hostname': 'juju-4665be-20180716142533-8.project.serverstack',
            'private-address': '10.5.0.17',
        }
        ssh_migrations.ssh_compute_add(
            UNIT1_PUBKEY_1,
            'nova-compute-lxd',
            rid='cloud-compute:23',
            unit='nova-compute-lxd/2')
        _ssh_compute_add_host_and_key.assert_called_once_with(
            UNIT1_PUBKEY_1,
            'juju-4665be-20180716142533-8.project.serverstack',
            '10.5.0.17',
            'nova-compute-lxd',
            user=None)

    @mock.patch.object(ssh_migrations, 'known_hosts')
    def test_ssh_known_hosts_lines(self, _known_hosts):
        _known_hosts.return_value = '/somedir/known_hosts'
        contents = '\n'.join(HOST_KEYS)
        with mock_open('/somedir/known_hosts', contents=contents):
            self.assertEqual(
                ssh_migrations.ssh_known_hosts_lines('nova-compute-lxd'),
                HOST_KEYS)

    @mock.patch.object(ssh_migrations, 'authorized_keys')
    def test_ssh_authorized_keys_lines(self, _authorized_keys):
        _authorized_keys.return_value = '/somedir/authorized_keys'
        contents = '\n'.join(PUB_KEYS)
        with mock_open('/somedir/authorized_keys', contents=contents):
            self.assertEqual(
                ssh_migrations.ssh_authorized_keys_lines('nova-compute-lxd'),
                PUB_KEYS)

    def setup_mocks_ssh_compute_remove(self, isfile, authorized_keys_lines):
        self.patch_object(
            ssh_migrations,
            'ssh_authorized_keys_lines',
            return_value=authorized_keys_lines)
        self.patch_object(ssh_migrations, 'known_hosts')
        self.patch_object(
            ssh_migrations,
            'authorized_keys',
            return_value='/somedir/authorized_keys')
        self.patch_object(
            ssh_migrations.os.path,
            'isfile',
            return_value=isfile)

    def test_ssh_compute_remove(self):
        self.setup_mocks_ssh_compute_remove(
            isfile=True,
            authorized_keys_lines=PUB_KEYS)
        with patch_open() as (mock_open, mock_file):
            ssh_migrations.ssh_compute_remove(
                UNIT1_PUBKEY_1,
                'nova-compute-lxd')
            mock_file.write.assert_called_with(UNIT2_PUBKEY_1 + '\n')
            mock_open.assert_called_with('/somedir/authorized_keys', 'w')

    def test_ssh_compute_remove_missing_file(self):
        self.setup_mocks_ssh_compute_remove(
            isfile=False,
            authorized_keys_lines=PUB_KEYS)
        with patch_open() as (mock_open, mock_file):
            ssh_migrations.ssh_compute_remove(
                UNIT1_PUBKEY_1,
                'nova-compute-lxd')
            self.assertFalse(mock_file.write.called)

    def test_ssh_compute_remove_missing_key(self):
        self.setup_mocks_ssh_compute_remove(
            isfile=False,
            authorized_keys_lines=[UNIT2_PUBKEY_1])
        with patch_open() as (mock_open, mock_file):
            ssh_migrations.ssh_compute_remove(
                UNIT1_PUBKEY_1,
                'nova-compute-lxd')
            self.assertFalse(mock_file.write.called)

    @mock.patch.object(ssh_migrations, 'ssh_known_hosts_lines')
    @mock.patch.object(ssh_migrations, 'ssh_authorized_keys_lines')
    def test_get_ssh_settings(self, _ssh_authorized_keys_lines,
                              _ssh_known_hosts_lines):
        _ssh_authorized_keys_lines.return_value = PUB_KEYS
        _ssh_known_hosts_lines.return_value = HOST_KEYS
        expect = {
            'known_hosts_0': UNIT1_HOST_KEY_1,
            'known_hosts_1': UNIT1_HOST_KEY_2,
            'known_hosts_2': UNIT2_HOST_KEY_1,
            'known_hosts_3': UNIT2_HOST_KEY_2,
            'known_hosts_max_index': 4,
            'authorized_keys_0': UNIT1_PUBKEY_1,
            'authorized_keys_1': UNIT2_PUBKEY_1,
            'authorized_keys_max_index': 2,
        }
        self.assertEqual(
            ssh_migrations.get_ssh_settings('nova-compute-lxd'),
            expect)

    @mock.patch.object(ssh_migrations, 'ssh_known_hosts_lines')
    @mock.patch.object(ssh_migrations, 'ssh_authorized_keys_lines')
    def test_get_ssh_settings_user(self, _ssh_authorized_keys_lines,
                                   _ssh_known_hosts_lines):
        _ssh_authorized_keys_lines.return_value = PUB_KEYS
        _ssh_known_hosts_lines.return_value = HOST_KEYS
        expect = {
            'nova_known_hosts_0': UNIT1_HOST_KEY_1,
            'nova_known_hosts_1': UNIT1_HOST_KEY_2,
            'nova_known_hosts_2': UNIT2_HOST_KEY_1,
            'nova_known_hosts_3': UNIT2_HOST_KEY_2,
            'nova_known_hosts_max_index': 4,
            'nova_authorized_keys_0': UNIT1_PUBKEY_1,
            'nova_authorized_keys_1': UNIT2_PUBKEY_1,
            'nova_authorized_keys_max_index': 2,
        }
        self.assertEqual(
            ssh_migrations.get_ssh_settings('nova-compute-lxd', user='nova'),
            expect)

    @mock.patch.object(ssh_migrations, 'ssh_known_hosts_lines')
    @mock.patch.object(ssh_migrations, 'ssh_authorized_keys_lines')
    def test_get_ssh_settings_empty(self, _ssh_authorized_keys_lines,
                                    _ssh_known_hosts_lines):
        _ssh_authorized_keys_lines.return_value = []
        _ssh_known_hosts_lines.return_value = []
        self.assertEqual(
            ssh_migrations.get_ssh_settings('nova-compute-lxd'),
            {})

    @mock.patch.object(ssh_migrations, 'get_ssh_settings')
    def test_get_all_user_ssh_settings(self, _get_ssh_settings):
        def ssh_setiings(application_name, user=None):
            base_settings = {
                'known_hosts_0': UNIT1_HOST_KEY_1,
                'known_hosts_max_index': 1,
                'authorized_keys_0': UNIT1_PUBKEY_1,
                'authorized_keys_max_index': 1}
            user_settings = {
                'nova_known_hosts_0': UNIT1_HOST_KEY_1,
                'nova_known_hosts_max_index': 1,
                'nova_authorized_keys_0': UNIT1_PUBKEY_1,
                'nova_authorized_keys_max_index': 1}
            if user:
                return user_settings
            else:
                return base_settings
        _get_ssh_settings.side_effect = ssh_setiings
        expect = {
            'known_hosts_0': UNIT1_HOST_KEY_1,
            'known_hosts_max_index': 1,
            'authorized_keys_0': UNIT1_PUBKEY_1,
            'authorized_keys_max_index': 1,
            'nova_known_hosts_0': UNIT1_HOST_KEY_1,
            'nova_known_hosts_max_index': 1,
            'nova_authorized_keys_0': UNIT1_PUBKEY_1,
            'nova_authorized_keys_max_index': 1}
        self.assertEqual(
            ssh_migrations.get_all_user_ssh_settings('nova-compute-lxd'),
            expect)
